#  Copyright (C) 2016 The Gvsbuild Authors
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, see <http://www.gnu.org/licenses/>.

"""Various downloader / unpacker (tar, git, ...)"""

import hashlib
import os
import shutil
import tarfile
import zipfile

from .simple_ui import log
from .utils import rmtree_full


def read_mark_file(directory, file_name=".wingtk-extracted-file"):
    """Read a single line from file, returning an empty string on error."""
    rt = ""
    try:
        with open(os.path.join(directory, file_name), encoding="utf-8") as fi:
            rt = fi.readline().strip()
    except OSError as e:
        log.debug(f"Exception on reading from '{file_name}'")
        log.debug(f"{e}")

    return rt


def write_mark_file(directory, val, file_name=".wingtk-extracted-file"):
    """Write the value (filename or content hash) to the mark file."""
    with open(os.path.join(directory, file_name), "w", encoding="utf-8") as fo:
        fo.write(f"{val}\n")


def extract_exec(
    src,
    dest_dir,
    dir_part=None,
    strip_one=False,
    check_file=None,
    force_dest=None,
    check_mark=False,
):
    """Extract (or copy, in case of an exe file) from src to dest_dir, handling
    the strip of the first part of the path in case of the tarbombs.

    dir_part is a piece, present in the tar/zip file, added to the dest_dir for
    the checks

    if check_file is passed and is present in the filesystem the extract is
    skipped (tool already installed)

    force_dest can be used only on the exe file and set the destination name

    with check_mark the name of the original file extracted is written in the
    destination dir and checked, forcing a new, clean, extraction

    Returns True if the extraction has been done, False if it's skipped (so we can skip
    marking the dependents of the project/tool)
    """

    # Support function
    def __get_stripped_tar_members(tar):
        for tarinfo in tar.getmembers():
            path = tarinfo.name.split("/")
            if len(path) == 1:
                if tarinfo.isdir():
                    continue
                else:
                    raise NotADirectoryError(
                        "Cannot strip directory prefix from tar with top level files"
                    )
            tarinfo.name = "/".join(path[1:])
            if tarinfo.issym() or tarinfo.islnk():
                tarinfo.linkname = "/".join(tarinfo.linkname.split("/")[1:])
            yield tarinfo

    full_dest = os.path.join(dest_dir, dir_part) if dir_part else dest_dir
    if check_mark:
        rd_file = read_mark_file(full_dest)
        wr_file = os.path.basename(src)
        if rd_file != wr_file:
            log.log(f"Forcing extraction of {src}")
            rmtree_full(full_dest, retry=True)
            check_file = None
        else:
            # ok, finish, we've done
            return False

    if check_file is not None:
        if check_file:
            # look for the specific file
            if os.path.isfile(check_file):
                log.debug(f"Skipping {src} handling, {check_file} present")
                return False
        else:
            # If the directory exist we are ok
            if os.path.exists(full_dest):
                log.debug(f"Skipping {src} handling, directory exists")
                return False

    log.log(f"Extracting {src} to {full_dest}")
    os.makedirs(full_dest, exist_ok=True)

    _n, ext = os.path.splitext(src.lower())
    if ext == ".exe":
        # Exe file, copy directly
        if force_dest:
            shutil.copy2(src, force_dest)
        else:
            shutil.copy2(src, dest_dir)
    elif ext == ".zip":
        # Zip file
        with zipfile.ZipFile(src) as zf:
            if strip_one:
                members = zf.infolist()
                for m in members:
                    if m.is_dir():
                        continue
                    cl = m.filename.split("/")
                    if len(cl) > 1:
                        m.filename = "/".join(cl[1:])

                    zf.extract(m, path=dest_dir)
            else:
                zf.extractall(path=dest_dir)
    else:
        # Ok, hoping it's a tarfile we can handle :)
        with tarfile.open(src) as tar:
            tar.extractall(
                dest_dir,
                __get_stripped_tar_members(tar) if strip_one else tar.getmembers(),
            )

    if check_mark:
        # write the data
        write_mark_file(full_dest, wr_file)

    # Say that we have done the extraction
    return True


def dirlist2set(st_dir, add_dirs=False, skipped_dir=None):
    """Loads & return a set with all the files and, eventually, directory from
    a single dir.

    Used to make a file list to create a .zip file
    """

    def _load_single_dir(dir_name, returned_set, skipped_dir):
        for cf in os.scandir(dir_name):
            full = os.path.join(dir_name, cf.name.lower())
            if cf.is_file():
                returned_set.add(full)
            elif cf.is_dir():
                if cf.name.lower() in skipped_dir:
                    log.debug(f"  Skipped dir '{cf.name}' (from '{dir_name}')")
                else:
                    if add_dirs:
                        returned_set.add(full)
                    _load_single_dir(full, returned_set, skipped_dir)

    rt = set()
    if skipped_dir is None:
        skipped_dir = []
    skipped_dir.append("__pycache__")
    try:
        log.debug(f"Getting file list from '{st_dir}'")
        _load_single_dir(st_dir, rt, set(skipped_dir))
    except FileNotFoundError:
        print(f"Warning: (--zip-continue) No file found on '{st_dir}'")
    return rt


def make_zip_hash(files):
    """Calculate an hash of all the files to put in a zip file."""
    hash_calc = hashlib.sha256()
    for file_name in sorted(files):
        # add also the file full path, to support only moving files in the zip
        hash_calc.update(file_name.lower().encode("utf-8"))
        if os.path.isfile(file_name):
            # add the file content
            with open(file_name, "rb") as fi:
                for chunk in iter(lambda: fi.read(4096), b""):
                    hash_calc.update(chunk)

    return hash_calc.hexdigest()


def make_zip(name, files, skip_spc=0):
    """Create the name .zip using all files.

    skip_spc spaces are dropped from the beginning of all file/dir names
    to avoid to have the full path (e.g. from
    c:\\data\temp\build\\my_arch we want to save only mt_arch
    """
    log.start_verbose(f"Creating zip file {name} with {len(files)} files")
    with zipfile.ZipFile(f"{name}.zip", "w", compression=zipfile.ZIP_DEFLATED) as zf:
        for f in sorted(files):
            zf.write(f, arcname=f[skip_spc:])
    log.end()


class Tarball:
    def update_build_dir(self):
        log.start_verbose(f"(tar) Updating {self.archive_file}")
        rt = extract_exec(
            self.archive_file,
            self.build_dir,
            strip_one=not self.tarbomb,
            check_mark=True,
        )
        log.end()
        return rt

    def unpack(self):
        log.start_verbose(f"(tar) Extracting {self.archive_file}")
        extract_exec(
            self.archive_file,
            self.build_dir,
            strip_one=not self.tarbomb,
            check_mark=True,
        )
        log.end()

    def export(self):
        log.start(f"(tar) Exporting {self.name}")

        path = os.path.join(self.export_dir, f"{self.name}.zip")
        with zipfile.ZipFile(path, "w") as zipped_path:
            log.log(f"(tar) Exporting {self.archive_file}")
            zipped_path.write(
                self.archive_file, arcname=os.path.basename(self.archive_file)
            )

            for p in self.patches:
                log.log(f"(tar) Exporting {p}")
                zipped_path.write(
                    os.path.join(self.build_dir, p),
                    arcname=f"patches/{os.path.basename(p)}",
                )

        log.end()


class GitRepo:
    def read_temp_hash(self):
        return read_mark_file(self.opts.git_expand_dir, f"{self.name}.hash")

    def write_temp_hash(self, hash_val):
        write_mark_file(self.opts.git_expand_dir, hash_val, f"{self.name}.hash")

    def get_tag_name(self, src_dir):
        if self.tag:
            # name the .zip from the tag, validating it
            t_name = [c if c.isalnum() else "_" for c in self.tag]
            tag_name = "".join(t_name)
        else:
            of = os.path.join(src_dir, ".git-temp.rsp")
            self.builder.exec_msys(
                f"git rev-parse --short HEAD >{of}", working_dir=src_dir
            )
            with open(of, encoding="utf-8") as fi:
                tag_name = fi.readline().rstrip("\n")
            os.remove(of)

        return tag_name

    def create_zip(self):
        """Create a .zip file with the git checkout to be able to work offline
        and as a reference of the last correct build."""
        src_dir = os.path.join(self.opts.git_expand_dir, self.name)
        zip_post = self.get_tag_name(src_dir)

        # Be sure to have the git .zip dir
        git_tmp_dir = os.path.join(self.builder.opts.archives_download_dir, "git")
        if not os.path.exists(git_tmp_dir):
            log.log(f"Creating git archives save directory {git_tmp_dir}")
            os.makedirs(git_tmp_dir)

        # check if some file has changed
        all_files = dirlist2set(
            src_dir,
            add_dirs=True,
            skipped_dir=[
                ".git",
            ],
        )
        n_hash = make_zip_hash(all_files)
        o_hash = None if self.clean else self.read_temp_hash()
        upd_build_dir = False
        if o_hash != n_hash:
            # create a .zip file with the downloaded project
            make_zip(
                os.path.join(git_tmp_dir, f"{self.prj_dir}-{zip_post}"),
                all_files,
                len(src_dir),
            )

            # update the hash
            self.write_temp_hash(n_hash)
            # copy the git buffer to the working dir, cleaning up before copying
            upd_build_dir = True
            if os.path.isdir(self.build_dir):
                rmtree_full(self.build_dir)

        if not upd_build_dir:
            if os.path.isdir(self.build_dir):
                o_hash = read_mark_file(self.build_dir)
                if o_hash != n_hash:
                    upd_build_dir = True
                    rmtree_full(self.build_dir)

            else:
                upd_build_dir = True
        if upd_build_dir:
            shutil.copytree(src_dir, self.build_dir)
            write_mark_file(self.build_dir, n_hash)

        return upd_build_dir

    def unpack(self):
        self.update_build_dir()

    def _update_dir(self, remove_dest=False):
        dest = os.path.join(self.opts.git_expand_dir, self.name)
        if (self.clean or remove_dest) and os.path.isdir(dest):
            rmtree_full(dest)

        if not os.path.isdir(dest):
            return self._clone_and_checkout(dest)
        # Update
        log.start(f"(git) Updating directory {dest}")

        if self.tag:
            self.builder.exec_msys("git fetch origin", working_dir=dest)
            self.builder.exec_msys(f"git checkout -f {self.tag}", working_dir=dest)
        else:
            self.builder.exec_msys("git checkout -f", working_dir=dest)
            self.builder.exec_msys("git pull --rebase", working_dir=dest)

        if self.fetch_submodules:
            self._update_submodules("Update submodule(s)", dest)
        return self.create_zip()

    def _clone_and_checkout(self, dest):
        log.start(f"(git) Cloning {self.repo_url} to {dest}")

        self.builder.exec_msys(f"git clone {self.repo_url} {dest}")

        if self.tag:
            self.builder.exec_msys(f"git checkout -f {self.tag}", working_dir=dest)

        if self.fetch_submodules:
            self._update_submodules("Fetch submodule(s)", dest)
        self.create_zip()
        return True

    def _update_submodules(self, log_value, dest):
        log.start_verbose(log_value)
        self.builder.exec_msys("git submodule update --init", working_dir=dest)
        log.end()

    def update_build_dir(self):
        rt = None
        if not os.path.exists(self.opts.git_expand_dir):
            log.log(f"Creating git expand directory {self.opts.git_expand_dir}")
            os.makedirs(self.opts.git_expand_dir, exist_ok=True)

        try:
            rt = self._update_dir()
        except Exception as e:
            log.message(f"{self.name}:Exception {e}")
            log.message("Removing the destination dir ...")
            rt = self._update_dir(remove_dest=True)

        if rt and os.path.exists(self.patch_dir):
            log.log(f"Copying files from {self.patch_dir} to {self.build_dir}")
            self.builder.copy_all(self.patch_dir, self.build_dir)
        log.end()
        return rt

    def export(self):
        log.start(f"(git) Exporting directory {self.build_dir}")

        src_dir = os.path.join(self.opts.git_expand_dir, self.name)
        filename = f"{self.name}-{self.get_tag_name(src_dir)}.zip"
        self.builder.exec_msys(
            f"git archive -o {filename} HEAD", working_dir=self.build_dir
        )

        path = os.path.join(self.export_dir, f"{self.name}.zip")
        with zipfile.ZipFile(path, "w") as zipped_path:
            log.log(f"(git) Exporting {filename}")
            zipped_path.write(os.path.join(self.build_dir, filename), arcname=filename)

            for p in self.patches:
                log.log(f"(git) Exporting {p}")
                zipped_path.write(
                    os.path.join(self.build_dir, p),
                    arcname=f"patches/{os.path.basename(p)}",
                )

        log.end()


class NullExpander:
    """Null expander to use when all the source are present in the script and
    nothing must be downloaded."""

    def update_build_dir(self):
        # Force the copy of the files in the script
        return True

    def unpack(self):
        # Everything is in our script, nothing to download
        pass

    def export(self):
        pass



#  Copyright (C) 2016 The Gvsbuild Authors
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, see <http://www.gnu.org/licenses/>.

from gvsbuild.utils.base_builders import CmakeProject
from gvsbuild.utils.base_expanders import Tarball
from gvsbuild.utils.base_project import Project, project_add


@project_add
class LevelDB(Tarball, CmakeProject):
    def __init__(self):
        Project.__init__(
            self,
            "leveldb",
            version="1.23",
            archive_url="https://github.com/google/leveldb/archive/tags/{version}.tar.gz",
            archive_filename="leveldb-{version}.tar.gz",
            hash="fa183b494f3ffe418a6804a57dfc0670080fbb0bcd1949cc4527d9b0077887a8",
            dependencies=[
                "cmake",
            ],
        )

    def build(self):
        CmakeProject.build(
            self,
            use_ninja=True,
            cmake_params="-DLEVELDB_BUILD_TESTS=OFF -DLEVELDB_BUILD_BENCHMARKS=OFF",
        )

        self.install(r".\LICENSE share\doc\leveldb")



#  Copyright (C) 2016 The Gvsbuild Authors
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, see <http://www.gnu.org/licenses/>.
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, see <http://www.gnu.org/licenses/>.

import os

from gvsbuild.utils.base_builders import Meson
from gvsbuild.utils.base_expanders import Tarball
from gvsbuild.utils.base_project import Project, project_add


@project_add
class GStreamer(Tarball, Meson):
    def __init__(self):
        Project.__init__(
            self,
            "gstreamer",
            repository="https://gitlab.freedesktop.org/gstreamer/gstreamer",
            version="1.24.1",
            lastversion_even=True,
            archive_url="https://gstreamer.freedesktop.org/src/gstreamer/gstreamer-{version}.tar.xz",
            hash="88fc3629a8cacf7955d8671b2ae867d4a52a07b6948daa25b67feed22f8055ef",
            dependencies=["meson", "ninja", "glib", "orc"],
        )

        if self.opts.enable_gi:
            self.add_dependency("gobject-introspection")
            enable_gi = "enabled"
        else:
            enable_gi = "disabled"

        self.add_param(f"-Dintrospection={enable_gi}")

    def build(self):
        add_path = os.path.join(self.builder.opts.msys_dir, "usr", "bin")

        Meson.build(
            self, add_path=add_path, meson_params="-Dtests=disabled -Dexamples=disabled"
        )
        self.install(r".\COPYING share\doc\gstreamer")


@project_add
class Orc(Tarball, Meson):
    def __init__(self):
        Project.__init__(
            self,
            "orc",
            version="0.4.38",
            lastversion_even=True,
            repository="https://gitlab.freedesktop.org/gstreamer/orc",
            archive_url="https://gstreamer.freedesktop.org/src/orc/orc-{version}.tar.xz",
            hash="a55a98d4772567aa3faed8fb84d540c3db77eaba16d3e2e10b044fbc9228668d",
            dependencies=["meson", "ninja"],
        )

    def build(self):
        Meson.build(self, meson_params="-Dbenchmarks=disabled -Dtools=enabled")
        self.install(r"COPYING share\doc\orc")


@project_add
class GstPluginsBase(Tarball, Meson):
    def __init__(self):
        Project.__init__(
            self,
            "gst-plugins-base",
            repository="https://gitlab.freedesktop.org/gstreamer/gstreamer",
            version="1.24.1",
            lastversion_even=True,
            archive_url="https://gstreamer.freedesktop.org/src/gst-plugins-base/gst-plugins-base-{version}.tar.xz",
            hash="884045d1d7c5d6bb8605e45c7ee0e9f1341888e81c2b7c42dff52bb98ede8ec3",
            dependencies=[
                "meson",
                "ninja",
                "gstreamer",
                "opus",
                "ogg",
            ],
        )
        # Examples depend on GTK3
        self.add_param("-Dexamples=disabled")

        if self.opts.enable_gi:
            self.add_dependency("gobject-introspection")
            enable_gi = "enabled"
        else:
            enable_gi = "disabled"

        self.add_param(f"-Dintrospection={enable_gi}")

    def build(self):
        Meson.build(
            self, meson_params=f'-Dc_link_args="{self.builder.gtk_dir}\\lib\\ogg.lib"'
        )
        self.install(r".\COPYING share\doc\gst-plugins-base")


@project_add
class GstPluginsGood(Tarball, Meson):
    def __init__(self):
        Project.__init__(
            self,
            "gst-plugins-good",
            repository="https://gitlab.freedesktop.org/gstreamer/gstreamer",
            version="1.24.1",
            lastversion_even=True,
            archive_url="https://gstreamer.freedesktop.org/src/gst-plugins-good/gst-plugins-good-{version}.tar.xz",
            hash="c02ba5583863e0bd9e265e5e4a27a03a3c473eb9ec1c10b07238f7176d9368a2",
            dependencies=[
                "meson",
                "ninja",
                "gst-plugins-base",
                "libvpx",
            ],
        )

    def build(self):
        Meson.build(self)
        self.install(r".\COPYING share\doc\gst-plugins-good")


@project_add
class GstPluginsBad(Tarball, Meson):
    def __init__(self):
        Project.__init__(
            self,
            "gst-plugins-bad",
            repository="https://gitlab.freedesktop.org/gstreamer/gstreamer",
            version="1.24.1",
            lastversion_even=True,
            archive_url="https://gstreamer.freedesktop.org/src/gst-plugins-bad/gst-plugins-bad-{version}.tar.xz",
            hash="f1a183966ea9136a00f159ee42db4d1d106abef022d1b49e63fba2fcbe50c8f8",
            dependencies=["meson", "ninja", "gst-plugins-base"],
            patches=[
                "wasapisink-reduce-buffer-latency.patch",
            ],
        )
        self.add_param("-Dcurl=disabled")
        self.add_param("-Dcurl-ssh2=disabled")

        if self.opts.enable_gi:
            self.add_dependency("gobject-introspection")
            enable_gi = "enabled"
        else:
            enable_gi = "disabled"

        self.add_param(f"-Dintrospection={enable_gi}")

    def build(self):
        Meson.build(self)
        self.install(r".\COPYING share\doc\gst-plugins-bad")


@project_add
class GstPluginsUgly(Tarball, Meson):
    def __init__(self):
        Project.__init__(
            self,
            "gst-plugins-ugly",
            repository="https://gitlab.freedesktop.org/gstreamer/gstreamer",
            version="1.24.1",
            lastversion_even=True,
            archive_url="https://gstreamer.freedesktop.org/src/gst-plugins-ugly/gst-plugins-ugly-{version}.tar.xz",
            hash="6462d36ce64b74203a8b302c109e762b1dd404a6235dfec3c4255106b7b86976",
            dependencies=["meson", "ninja", "gst-plugins-base"],
        )

    def build(self):
        Meson.build(self)
        self.install(r".\COPYING share\doc\gst-plugins-ugly")


@project_add
class GstDevTools(Tarball, Meson):
    def __init__(self):
        Project.__init__(
            self,
            "gst-devtools",
            repository="https://gitlab.freedesktop.org/gstreamer/gstreamer",
            version="1.24.1",
            lastversion_even=True,
            archive_url="https://gstreamer.freedesktop.org/src/gst-devtools/gst-devtools-{version}.tar.xz",
            hash="49ae0347f463641b094ebbc63350936a29e8a063ca90cd0b44a49e0c2b2129d2",
            dependencies=["meson", "ninja", "json-glib"],
        )

        if self.opts.enable_gi:
            self.add_dependency("gobject-introspection")
            enable_gi = "enabled"
        else:
            enable_gi = "disabled"

        self.add_param(f"-Dintrospection={enable_gi}")

    def build(self):
        Meson.build(self)
        self.install(r".\COPYING share\doc\gst-devtools")


@project_add
class GstPython(Tarball, Meson):
    def __init__(self):
        Project.__init__(
            self,
            "gst-python",
            repository="https://gitlab.freedesktop.org/gstreamer/gstreamer",
            version="1.24.1",
            lastversion_even=True,
            archive_url="https://gstreamer.freedesktop.org/src/gst-python/gst-python-{version}.tar.xz",
            hash="5a23732783edae03cc2164c6927c05477bbb3a7008f8939291b855473781a651",
            dependencies=["meson", "ninja", "pygobject", "gst-plugins-base"],
            patches=[
                "001-fix-instsoname-not-found.patch",
            ],
        )

    def build(self):
        Meson.build(self)
        self.install(r".\COPYING share\doc\gst-python")



#  Copyright (C) 2016 The Gvsbuild Authors
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, see <http://www.gnu.org/licenses/>.

from gvsbuild.utils.base_builders import CmakeProject
from gvsbuild.utils.base_expanders import Tarball
from gvsbuild.utils.base_project import Project, project_add


@project_add
class FreeRDP(Tarball, CmakeProject):
    def __init__(self):
        Project.__init__(
            self,
            "freerdp",
            version="2.11.2",
            archive_url="https://github.com/FreeRDP/FreeRDP/files/12670558/freerdp-{version}.tar.gz",
            hash="17e36553065cb59126f8fee987ec74783db50232b361574dac86b7a768dea3b4",
            dependencies=[
                "cmake",
                "ninja",
                "openssl",
                "openh264",
                "ffmpeg",
                "x264",
                "zlib",
                "libjpeg-turbo",
            ],
        )

    def build(self):
        CmakeProject.build(
            self, use_ninja=True, cmake_params="-DWITH_SSE2=ON -DCHANNEL_URBDRC=OFF"
        )

        self.install(r".\LICENSE share\doc\freerdp")



import typer


def version_callback(active: bool):
    if not active:
        return

    import importlib.metadata

    version = importlib.metadata.version("gvsbuild")
    typer.echo(f"gvsbuild v{version}")
    raise typer.Exit()



#  Copyright (C) 2016 The Gvsbuild Authors
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, see <http://www.gnu.org/licenses/>.

from gvsbuild.utils.base_builders import Meson
from gvsbuild.utils.base_expanders import Tarball
from gvsbuild.utils.base_project import Project, project_add


@project_add
class Libsoup2(Tarball, Meson):
    def __init__(self):
        Project.__init__(
            self,
            "libsoup2",
            lastversion_major=2,
            repository="https://gitlab.gnome.org/GNOME/libsoup",
            version="2.74.3",
            lastversion_even=True,
            archive_url="https://download.gnome.org/sources/libsoup/{major}.{minor}/libsoup-{version}.tar.xz",
            hash="e4b77c41cfc4c8c5a035fcdc320c7bc6cfb75ef7c5a034153df1413fa1d92f13",
            dependencies=[
                "libxml2",
                "glib-networking",
                "sqlite",
                "libpsl",
                "mit-kerberos",
            ],
        )

        if self.opts.enable_gi:
            self.add_dependency("gobject-introspection")
            enable_gi = "enabled"
        else:
            enable_gi = "disabled"

        self.add_param(f"-Dintrospection={enable_gi}")
        self.add_param("-Dvapi=disabled")
        self.add_param("-Dsysprof=disabled")
        self.add_param("-Dtls_check=false")
        self.add_param("-Dtests=false")

    def build(self):
        Meson.build(self)

        self.install(r".\COPYING share\doc\libsoup2")


@project_add
class Libsoup3(Tarball, Meson):
    def __init__(self):
        Project.__init__(
            self,
            "libsoup3",
            version="3.4.4",
            lastversion_major=3,
            lastversion_even=True,
            repository="https://gitlab.gnome.org/GNOME/libsoup",
            archive_url="https://download.gnome.org/sources/libsoup/{major}.{minor}/libsoup-{version}.tar.xz",
            hash="291c67725f36ed90ea43efff25064b69c5a2d1981488477c05c481a3b4b0c5aa",
            dependencies=[
                "libxml2",
                "glib-networking",
                "sqlite",
                "libpsl",
                "mit-kerberos",
                "nghttp2",
            ],
        )

        if self.opts.enable_gi:
            self.add_dependency("gobject-introspection")
            enable_gi = "enabled"
        else:
            enable_gi = "disabled"

        self.add_param(f"-Dintrospection={enable_gi}")
        self.add_param("-Dvapi=disabled")
        self.add_param("-Dsysprof=disabled")
        self.add_param("-Dtls_check=false")
        self.add_param("-Dtests=false")

    def build(self):
        Meson.build(self)

        self.install(r".\COPYING share\doc\libsoup3")



#  Copyright (C) 2016 The Gvsbuild Authors
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, see <http://www.gnu.org/licenses/>.
import sys
from pathlib import Path

from gvsbuild.utils.base_builders import Meson
from gvsbuild.utils.base_expanders import Tarball
from gvsbuild.utils.base_project import Project, project_add
from gvsbuild.utils.simple_ui import log
from gvsbuild.utils.utils import python_find_libs_dir


@project_add
class GObjectIntrospection(Tarball, Meson):
    def __init__(self):
        Project.__init__(
            self,
            "gobject-introspection",
            version="1.80.1",
            lastversion_even=True,
            repository="https://gitlab.gnome.org/GNOME/gobject-introspection",
            archive_url="https://download.gnome.org/sources/gobject-introspection/{major}.{minor}/gobject-introspection-{version}.tar.xz",
            hash="a1df7c424e15bda1ab639c00e9051b9adf5cea1a9e512f8a603b53cd199bc6d8",
            dependencies=[
                "ninja",
                "meson",
                "msys2",
                "pkgconf",
                "glib-base",
            ],
            patches=[
                # https://gitlab.gnome.org/GNOME/gobject-introspection/-/issues/427
                "001-incorrect-giscanner-path.patch",
            ],
        )

    def build(self):
        # For finding gobject-introspection.pc
        self.builder.mod_env("PKG_CONFIG_PATH", ".")
        # For finding & using girepository.lib/.dll
        self.builder.mod_env("LIB", r".\girepository")
        self.builder.mod_env("PATH", r".\girepository")
        # For linking the _giscanner.pyd extension module when using a virtualenv
        py_dir = Path(sys.executable).parent
        py_libs = python_find_libs_dir(py_dir)
        if py_libs:
            log.debug(f"Python library path is [{py_libs}]")
            self.builder.mod_env("LIB", py_libs, prepend=False)

        Meson.build(
            self,
            meson_params=f'-Dpython="{py_dir}\\python.exe" -Dcairo_libname=cairo-gobject-2.dll',
        )



#  Copyright (C) 2016 The Gvsbuild Authors
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, see <http://www.gnu.org/licenses/>.

"""gvsbuild deps print / .gv graph."""

from typing import List

import typer

# Verify we can import from the script directory
try:
    import gvsbuild.utils.utils
except ImportError:
    # We are probably using an embedded installation
    print(
        "Error importing utility (running the embedded interpreter ?), fixing paths ..."
    )
    import os
    import sys

    # Get the script dir
    script_dir = os.path.dirname(os.path.realpath(sys.argv[0]))
    # and add it at the beginning, emulating the standard python startup
    sys.path.insert(0, script_dir)

import gvsbuild.groups  # noqa: F401
import gvsbuild.projects  # noqa: F401
import gvsbuild.tools  # noqa: F401
from gvsbuild.utils.base_project import Project, ProjectType
from gvsbuild.utils.utils import ordered_set


def print_deps(flatten=False, add_all=False):
    done = []

    def dump_single_dep(st, name, flatten):
        if flatten:
            if not st:
                done.append(name)
        else:
            if st:
                # dependency
                print(f"{st}{name}")
            else:
                print(f"  > {name}")
                st = "   "
            done.append(name)

        rt = False
        p = Project._dict[name]
        if p.dependencies:
            for d in p.dependencies:
                add = True
                if not add_all:
                    ty = Project._dict[d].type
                    if ty != ProjectType.PROJECT:
                        add = False

                if add:
                    rt = True
                    if d in done:
                        if not flatten:
                            print(f"{st}    {d} *")
                    else:
                        done.append(d)
                        dump_single_dep(f"{st}    ", d, flatten)
        return rt

    prj = [x.name for x in Project._projects if x.type == ProjectType.PROJECT]
    print("Projects dependencies:")
    for n in prj:
        done = []
        if flatten:
            print(f"> {n}")
        if dump_single_dep("", n, flatten):
            if flatten:
                done.remove(n)
                for t in sorted(done):
                    print(f"    {t}")

            else:
                print("")


def make_graph(
    out_file,
    put_all=False,
    invert_dep=False,
    add_tools=False,
    add_groups=False,
    skip=None,
):
    gr_colors = [
        0x000080,
        0x008000,
        0x008080,
        0x800000,
        0x800080,
        0x808000,
        0x808080,
        0x0000F0,
        0x00F000,
        0x00F0F0,
        0xF00000,
        0xF000F0,
        0xF0F000,
        0xF00080,
        0xF08000,
        0xF08080,
        0x80F000,
        0x80F080,
        0x00F080,
        0x0080F0,
        0x8000F0,
        0x8080F0,
    ]
    gr_index = 0

    to_skip = set(skip)
    with open(out_file, "w", encoding="utf-8") as fo:
        print(f"Writing file {out_file}")
        used = set()
        fo.write("digraph gtk3dep {\n")
        for n in Project._names:
            if n not in to_skip:
                t = Project._dict[n]

                add = True
                if t.type == ProjectType.TOOL:
                    add = add_tools
                elif t.type == ProjectType.GROUP:
                    add = add_groups
                else:
                    add = True

                if add:
                    if t.dependencies:
                        gr_index += 1
                        gr_index %= len(gr_colors)
                        for d in t.dependencies:
                            if d in to_skip:
                                print(f"Skip '{d}' for '{n}'")
                            else:
                                if invert_dep:
                                    fo.write(
                                        f'    "{d}" -> "{n}" [color="#{gr_colors[gr_index]:06x}"];\n'
                                    )
                                else:
                                    fo.write(
                                        f'    "{n}" -> "{d}" [color="#{gr_colors[gr_index]:06x}"];\n'
                                    )
                                used.add(d)
                else:
                    used.add(t.name)

        if put_all:
            # Puts all projects that are not referenced from others
            for n in Project._names:
                if n not in used:
                    fo.write(f'    "BUILD" -> "{n}" [color="#c00080"];\n')

        fo.write("};\n")


def compute_deps(proj):
    if hasattr(proj, "all_dependencies"):
        return
    dependencies = ordered_set()
    for dep in proj.dependencies:
        compute_deps(dep)
        for p in dep.all_dependencies:
            dependencies.add(p)
        dependencies.add(dep)
    proj.all_dependencies = dependencies


def deps(
    flatten: bool = typer.Option(False, help="Flatten the dependencies"),
    dep_tools: bool = typer.Option(
        False,
        help="Include tools in the dependencies",
    ),
    graph: bool = typer.Option(
        False, help="Generate a graphviz file", rich_help_panel="Graphing Options"
    ),
    graph_all: bool = typer.Option(
        False,
        help="Also include unreferenced projects to the graph",
        rich_help_panel="Graphing Options",
    ),
    add_tools: bool = typer.Option(
        False, help="Include tools in the graph", rich_help_panel="Graphing Options"
    ),
    add_groups: bool = typer.Option(
        False,
        help="Include group projects in the graph",
        rich_help_panel="Graphing Options",
    ),
    gv_file: str = typer.Option(
        "wingtk.gv", help="Graphviz output file", rich_help_panel="Graphing Options"
    ),
    invert: bool = typer.Option(
        False, help="Invert the dependencies", rich_help_panel="Graphing Options"
    ),
    skip: List[str] = typer.Option(
        None,
        help="A comma separated list of projects not to graph",
        rich_help_panel="Graphing Options",
    ),
):
    Project.add_all()
    # do what's asked
    if graph:
        # .gv graph
        make_graph(
            out_file=gv_file,
            put_all=graph_all,
            invert_dep=invert,
            add_tools=add_tools,
            add_groups=add_groups,
            skip=skip,
        )
    else:
        # simple dep print
        print_deps(flatten=flatten, add_all=dep_tools)



#  Copyright (C) 2016 The Gvsbuild Authors
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, see <http://www.gnu.org/licenses/>.

import json
import sys
from typing import List

import typer
from packaging import version

from gvsbuild.utils.base_project import Project, ProjectType


def set_projects_latest_versions(projects):
    try:
        import lastversion
    except ImportError:
        print("Please pip install lastversion in your Python environment")
        sys.exit(1)

    projects = {
        project.name: project
        for project in Project.list_projects()
        if project.type.value in [ProjectType.PROJECT, ProjectType.TOOL]
        and not project.internal
    }

    try:
        for project in projects.values():
            if project.internal:
                continue
            project.latest_version = lastversion.latest(
                repo=project.repository or project.name,
                major=project.lastversion_major,
            )
            if project.latest_version:
                project.outdated = version.parse(
                    str(project.latest_version)
                ) > version.parse(project.version)

    except lastversion.exceptions.ApiCredentialsError:
        print(
            "Create or update the GITHUB token at https://github.com/settings/tokens, then set or update the token "
            "environmental variable with:\n$env:GITHUB_API_TOKEN=xxxxxxxxxxxxxxx"
        )
        exit(1)


def list_(
    projects_names: List[str] = typer.Argument(None, help="The projects to list"),
    project_type: ProjectType = typer.Option(
        None,
        "--type",
        help="Specify type of projects to show, if not selected show all",
        rich_help_panel="Selection Options",
    ),
    show_deps: bool = typer.Option(
        False,
        "--deps",
        help="Include dependencies, only useful when selecting the projects to show",
        rich_help_panel="Selection Options",
    ),
    json_: bool = typer.Option(
        False,
        "--json",
        help="Show list in JSON format",
        rich_help_panel="Formatting Options",
    ),
    latest: bool = typer.Option(
        False,
        "--latest",
        help="Fetch latest information",
        rich_help_panel="Formatting Options",
    ),
    outdated: bool = typer.Option(
        False,
        "--outdated",
        help="Only show outdated projects",
        rich_help_panel="Selection Options",
    ),
):
    Project.add_all()

    projects = Project.list_projects()
    if projects_names:
        projects = [project for project in projects if project.name in projects_names]
    if show_deps:
        projects = Project.compute_dependencies(projects)
    if project_type:
        projects = [
            project for project in projects if project.type.value == project_type
        ]

    if outdated:
        latest = True

    if latest:
        set_projects_latest_versions(projects)

    if outdated:
        projects = [
            project
            for project in projects
            if hasattr(project, "outdated") and project.outdated
        ]

    if json_:

        def _get_project_data(project):
            data = {"dependencies": project.dependencies, "type": project.type.value}
            if project.version:
                data["version"] = project.version
            if hasattr(project, "latest_version") and project.latest_version:
                data["latest-version"] = str(project.latest_version)
            return data

        print(
            json.dumps(
                {project.name: _get_project_data(project) for project in projects},
                indent=4,
                sort_keys=True,
            )
        )
    else:
        for project_type in ProjectType:
            type_projects = [
                project for project in projects if project.type == project_type
            ]
            if type_projects:
                print(f"Available projects with type {project_type}:")
                for project in type_projects:
                    params = {
                        "name": f"{project.name:<{Project.name_len}}",
                        "version": f"{project.version:<45}",
                    }
                    if latest:
                        if (
                            hasattr(project, "latest_version")
                            and project.latest_version
                        ):
                            params[
                                "latest_version"
                            ] = f"{str(project.latest_version):<45}"
                        else:
                            params["latest_version"] = "undefined"
                        print("\t{name} {version} {latest_version}".format(**params))
                    else:
                        print("\t{name} {version}".format(**params))



#  Copyright (C) 2016 The Gvsbuild Authors
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, see <http://www.gnu.org/licenses/>.

import sys

from packaging import version

from gvsbuild.utils.base_project import Project, ProjectType


def outdated():
    try:
        import lastversion
    except ImportError:
        print("Please pip install lastversion in your Python environment")
        sys.exit(0)

    Project.add_all()
    projects = {
        project.name: project
        for project in Project.list_projects()
        if project.type.value in [ProjectType.PROJECT, ProjectType.TOOL]
        and not project.internal
    }

    print("Looking for projects and tools that are out-of-date, please submit a PR!")
    print(f"\t{'Project Name':<{Project.name_len}} {'Current':<45} {'Latest':<45}")
    try:
        for project in projects.values():
            try:
                latest_version = lastversion.latest(
                    repo=project.repository or project.name,
                    major=project.lastversion_major,
                    even=project.lastversion_even,
                )

                if not latest_version:
                    print(
                        f"\t{project.name:<{Project.name_len}} {project.version:<45} {'No release found':<45}"
                    )
                elif version.parse(str(latest_version)) > version.parse(
                    project.version
                ):
                    print(
                        f"\t{project.name:<{Project.name_len}} {project.version:<45} {str(latest_version):<45}"
                    )
            except version.InvalidVersion:
                print(f"Project {project.name} does not have a valid version")
    except lastversion.exceptions.ApiCredentialsError:
        print(
            "Create or update the GITHUB token at https://github.com/settings/tokens, then set or update the token environmental variable with:\n$env:GITHUB_API_TOKEN=xxxxxxxxxxxxxxx"
        )



#  Copyright (C) 2016 The Gvsbuild Authors
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, see <http://www.gnu.org/licenses/>.

"""Default tools used to build the various projects."""

import os
import subprocess

from .utils.base_expanders import extract_exec
from .utils.base_tool import Tool, tool_add


@tool_add
class ToolCargo(Tool):
    def __init__(self):
        Tool.__init__(
            self,
            "cargo",
            version="1.77.1",
            repository="rust-lang/rust",
            archive_url="https://win.rustup.rs/x86_64",
            archive_filename="rustup-init.exe",
            exe_name="cargo.exe",
        )

    def load_defaults(self):
        Tool.load_defaults(self)
        self.tool_path = os.path.join(self.build_dir, "bin")
        self.full_exe = os.path.join(self.tool_path, "cargo.exe")

        self.add_extra_env("RUSTUP_HOME", self.build_dir)
        self.add_extra_env("CARGO_HOME", self.build_dir)

    def unpack(self):
        env = os.environ.copy()
        env["RUSTUP_HOME"] = self.build_dir
        env["CARGO_HOME"] = self.build_dir

        toolchain = (
            f'{self.version}-{"i686" if self.opts.x86 else "x86_64"}-pc-windows-msvc'
        )
        subprocess.run(
            f"{self.archive_file} --no-modify-path --default-toolchain {toolchain} -y",
            check=True,
            env=env,
        )

        self.mark_deps = True


@tool_add
class ToolCmake(Tool):
    def __init__(self):
        Tool.__init__(
            self,
            "cmake",
            version="3.29.0",
            archive_url="https://github.com/Kitware/CMake/releases/download/v{version}/cmake-{version}-windows-x86_64.zip",
            hash="9ab28eba1ab7911a0e57ab274f5990a283fffa1d22eb711792d5562e5869f9ef",
            dir_part="cmake-{version}-windows-x86_64",
        )

    def load_defaults(self):
        Tool.load_defaults(self)
        self.tool_path = os.path.join(self.build_dir, "bin")
        self.full_exe = os.path.join(self.tool_path, "cmake.exe")

    def unpack(self):
        self.mark_deps = extract_exec(
            self.archive_file,
            self.opts.tools_root_dir,
            dir_part=self.dir_part,
            check_file=self.full_exe,
            check_mark=True,
        )


@tool_add
class ToolMeson(Tool):
    def __init__(self):
        Tool.__init__(
            self,
            "meson",
            version="1.4.0",
            archive_url="https://github.com/mesonbuild/meson/archive/refs/tags/{version}.tar.gz",
            archive_filename="meson-{version}.tar.gz",
            hash="61382f295378bddcd9bebb3a9a9065b1cbc671fa41b80964ab02726f9a5f3a88",
            dir_part="meson-{version}",
            exe_name="meson.py",
        )

    def unpack(self):
        self.mark_deps = extract_exec(
            self.archive_file,
            self.builder.opts.tools_root_dir,
            dir_part=self.dir_part,
            check_file=self.full_exe,
            check_mark=True,
        )


@tool_add
class ToolMsys2(Tool):
    def __init__(self):
        Tool.__init__(self, "msys2")
        self.internal = True

    def load_defaults(self):
        Tool.load_defaults(self)
        self.tool_path = os.path.join(self.opts.msys_dir, "usr", "bin")

    def unpack(self):
        self.tool_mark()

    def get_path(self):
        # We always put msys at the end of path
        return None, self.tool_path


@tool_add
class ToolNasm(Tool):
    def __init__(self):
        Tool.__init__(
            self,
            "nasm",
            version="2.16.01",
            archive_url="https://www.nasm.us/pub/nasm/releasebuilds/{version}/win64/nasm-{version}-win64.zip",
            hash="029eed31faf0d2c5f95783294432cbea6c15bf633430f254bb3c1f195c67ca3a",
            dir_part="nasm-{version}",
            exe_name="nasm.exe",
        )

    def unpack(self):
        # We download directly the exe file so we copy it on the tool directory ...
        self.mark_deps = extract_exec(
            self.archive_file,
            self.builder.opts.tools_root_dir,
            dir_part=self.dir_part,
            check_file=self.full_exe,
            force_dest=self.full_exe,
            check_mark=True,
        )


@tool_add
class ToolNinja(Tool):
    def __init__(self):
        Tool.__init__(
            self,
            "ninja",
            version="1.11.1",
            archive_url="https://github.com/ninja-build/ninja/releases/download/v{version}/ninja-win.zip",
            archive_filename="ninja-win-{version}.zip",
            hash="524b344a1a9a55005eaf868d991e090ab8ce07fa109f1820d40e74642e289abc",
            dir_part="ninja-{version}",
            exe_name="ninja.exe",
        )

    def unpack(self):
        self.mark_deps = extract_exec(
            self.archive_file, self.build_dir, check_file=self.full_exe, check_mark=True
        )


@tool_add
class ToolPerl(Tool):
    def __init__(self):
        Tool.__init__(
            self,
            "perl",
            version="5.20.0",
            archive_url="https://github.com/wingtk/gtk-win32/releases/download/Perl-{major}.{minor}/perl-{version}-x64.tar.xz",
            hash="05e01cf30bb47d3938db6169299ed49271f91c1615aeee5649174f48ff418c55",
            dir_part="perl-{version}",
        )

    def load_defaults(self):
        Tool.load_defaults(self)
        # Set the builder object to point to the path to use, when we need to pass directly the executable to *make
        self.base_dir = os.path.join(self.build_dir, "x64")
        # full path, added to the environment when needed
        self.tool_path = os.path.join(self.base_dir, "bin")
        self.full_exe = os.path.join(self.tool_path, "perl.exe")

    def unpack(self):
        self.mark_deps = extract_exec(
            self.archive_file, self.build_dir, check_file=self.full_exe, check_mark=True
        )

    def get_base_dir(self):
        return self.base_dir


@tool_add
class ToolGo(Tool):
    def __init__(self):
        Tool.__init__(
            self,
            "go",
            version="1.22.1",
            archive_url="https://go.dev/dl/go{version}.windows-amd64.zip",
            hash="cf9c66a208a106402a527f5b956269ca506cfe535fc388e828d249ea88ed28ba",
            dir_part="go-{version}",
        )

    def load_defaults(self):
        Tool.load_defaults(self)
        self.tool_path = os.path.join(self.build_dir, "bin")
        self.full_exe = os.path.join(self.tool_path, "go.exe")

    def unpack(self):
        # We download directly the exe file, so we copy it to the tool directory
        self.mark_deps = extract_exec(
            self.archive_file,
            self.build_dir,
            check_file=self.full_exe,
            check_mark=True,
            strip_one=True,
        )



#  Copyright (C) 2016 The Gvsbuild Authors
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, see <http://www.gnu.org/licenses/>.

from gvsbuild.utils.base_builders import CmakeProject
from gvsbuild.utils.base_expanders import Tarball
from gvsbuild.utils.base_project import project_add


@project_add
class Opus(Tarball, CmakeProject):
    def __init__(self):
        CmakeProject.__init__(
            self,
            "opus",
            version="1.5.1",
            repository="https://github.com/xiph/opus",
            archive_url="https://downloads.xiph.org/releases/opus/opus-{version}.tar.gz",
            hash="b84610959b8d417b611aa12a22565e0a3732097c6389d19098d844543e340f85",
            dependencies=[
                "ninja",
                "cmake",
                "pkgconf",
            ],
        )

    def build(self):
        CmakeProject.build(
            self,
            use_ninja=True,
            cmake_params="-DOPUS_BUILD_SHARED_LIBRARY=ON -DOPUS_BUILD_TESTING=OFF",
        )
        self.install(r"COPYING share\doc\opus")

        configuration = (
            "debug-optimized"
            if self.opts.release_configuration_is_actually_debug_optimized
            else self.opts.configuration
        )
        if configuration in ["debug-optimized", "debug"]:
            self.install(r".\_gvsbuild-cmake\opus.pdb bin")

        # FIXME: remove once we switch back to meson
        self.install_pc_files()



#  Copyright (C) 2016 The Gvsbuild Authors
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, see <http://www.gnu.org/licenses/>.

from gvsbuild.utils.base_builders import Meson
from gvsbuild.utils.base_expanders import NullExpander, Tarball
from gvsbuild.utils.base_project import Project, project_add


@project_add
class GLibBase(Tarball, Meson):
    def __init__(self):
        Meson.__init__(
            self,
            "glib-base",
            version="2.80.0",
            lastversion_even=True,
            repository="https://gitlab.gnome.org/GNOME/glib",
            archive_url="https://download.gnome.org/sources/glib/{major}.{minor}/glib-{version}.tar.xz",
            hash="8228a92f92a412160b139ae68b6345bd28f24434a7b5af150ebe21ff587a561d",
            dependencies=[
                "ninja",
                "meson",
                "pkgconf",
                "gettext",
                "libffi",
                "zlib",
                "pcre2",
            ],
            patches=[
                "001-glib-package-installation-directory.patch",
                # https://gitlab.gnome.org/GNOME/gobject-introspection/-/issues/499
                "002-gir-scanner-dll-not-found.patch",
            ],
        )
        self.add_param("-Dman-pages=disabled")
        self.add_param("-Dtests=false")
        self.add_param("-Ddocumentation=false")
        self.add_param("-Dintrospection=disabled")

    def build(self):
        Meson.build(self)
        self.install(r".\LICENSES\* share\doc\glib")


@project_add
class GLib(Tarball, Meson):
    def __init__(self):
        Meson.__init__(
            self,
            "glib",
            version="2.80.0",
            lastversion_even=True,
            repository="https://gitlab.gnome.org/GNOME/glib",
            archive_url="https://download.gnome.org/sources/glib/{major}.{minor}/glib-{version}.tar.xz",
            hash="8228a92f92a412160b139ae68b6345bd28f24434a7b5af150ebe21ff587a561d",
            dependencies=["glib-base"],
            patches=[
                "001-glib-package-installation-directory.patch",
                # https://gitlab.gnome.org/GNOME/gobject-introspection/-/issues/499
                "002-gir-scanner-dll-not-found.patch",
            ],
        )
        self.add_param("-Dman-pages=disabled")
        self.add_param("-Dtests=false")
        self.add_param("-Ddocumentation=false")
        if self.opts.enable_gi:
            self.add_dependency("gobject-introspection")
            self.add_param("-Dintrospection=enabled")

    def build(self):
        if self.opts.enable_gi:
            Meson.build(self)


@project_add
class GLibNetworking(Tarball, Meson):
    def __init__(self):
        Project.__init__(
            self,
            "glib-networking",
            version="2.80.0",
            lastversion_even=True,
            repository="https://gitlab.gnome.org/GNOME/glib-networking",
            archive_url="https://download.gnome.org/sources/glib-networking/{major}.{minor}/glib-networking-{version}.tar.xz",
            hash="d8f4f1aab213179ae3351617b59dab5de6bcc9e785021eee178998ebd4bb3acf",
            dependencies=[
                "pkgconf",
                "ninja",
                "meson",
                "glib",
                "openssl",
                "gsettings-desktop-schemas",
            ],
        )

    def build(self):
        Meson.build(
            self, meson_params="-Dgnutls=disabled -Dopenssl=enabled -Dlibproxy=disabled"
        )
        self.install(r".\COPYING share\doc\glib-networking")
        self.install(r".\LICENSE_EXCEPTION share\doc\glib-networking")


@project_add
class GLibPyWrapper(NullExpander, Meson):
    def __init__(self):
        Project.__init__(
            self,
            "glib-py-wrapper",
            dependencies=["glib"],
            version="0.1.0",
            internal=True,
            repository="https://gitlab.gnome.org/GNOME/glib-py-wrapper",
        )

    def build(self):
        Meson.build(self)



#  Copyright (C) 2016 The Gvsbuild Authors
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, see <http://www.gnu.org/licenses/>.

from gvsbuild.utils.base_builders import Meson
from gvsbuild.utils.base_expanders import Tarball
from gvsbuild.utils.base_project import project_add


@project_add
class Pangomm(Tarball, Meson):
    def __init__(self):
        Meson.__init__(
            self,
            "pangomm",
            prj_dir="pangomm",
            version="2.52.0",
            lastversion_even=True,
            repository="https://gitlab.gnome.org/GNOME/pangomm",
            archive_url="https://download.gnome.org/sources/pangomm/{major}.{minor}/pangomm-{version}.tar.xz",
            hash="34a134126a6484ff12f774358c36ecc44d0e9df094e1b83796d9774bb7d24947",
            dependencies=[
                "meson",
                "ninja",
                "libsigc++",
                "cairomm",
                "pango",
                "glibmm",
            ],
        )

    def build(self):
        Meson.build(
            self,
            meson_params="-Dbuild-documentation=false",
        )

        self.install(r".\COPYING share\doc\glibmm")



#  Copyright (C) 2016 The Gvsbuild Authors
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, see <http://www.gnu.org/licenses/>.

from gvsbuild.utils.base_builders import Meson
from gvsbuild.utils.base_expanders import Tarball
from gvsbuild.utils.base_project import Project, project_add


@project_add
class Freetype(Tarball, Meson):
    def __init__(self):
        Project.__init__(
            self,
            "freetype",
            version="2.13.2",
            repository="https://gitlab.freedesktop.org/freetype/freetype",
            archive_url="https://gitlab.freedesktop.org/freetype/freetype/-/archive/VER-{major}-{minor}-{micro}/freetype-VER-{major}-{minor}-{micro}.tar.gz",
            hash="427201f5d5151670d05c1f5b45bef5dda1f2e7dd971ef54f0feaaa7ffd2ab90c",
            dependencies=["pkgconf", "ninja", "libpng"],
            patches=["0001-meson-in-shared-libraries-we-need-to-export-the-meth.patch"],
        )

    def build(self):
        Meson.build(self)
        self.install(r".\docs\LICENSE.TXT share\doc\freetype")



#  Copyright (C) 2016 The Gvsbuild Authors
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, see <http://www.gnu.org/licenses/>.

import contextlib

from gvsbuild.utils.base_expanders import Tarball
from gvsbuild.utils.base_project import Project, project_add


@project_add
class OpenSSLBase(Tarball, Project):
    def __init__(self):
        Project.__init__(
            self,
            "openssl-base",
            version="3.2.1",
            archive_url="https://www.openssl.org/source/openssl-{version}.tar.gz",
            hash="83c7329fe52c850677d75e5d0b0ca245309b97e8ecbcfdc1dfdc4ab9fac35b39",
            dependencies=[
                "perl",
                "nasm",
                "msys2",
            ],
            patches=["0001-crypto-providers-config.patch"],
        )

    def build(self):
        common_options = r"enable-fips no-comp no-docs no-ssl3 --openssldir=%(gtk_dir)s/etc/ssl --prefix=%(gtk_dir)s"
        debug_option = "debug-" if self.builder.opts.configuration == "debug" else ""
        target_option = "VC-WIN32 " if self.builder.x86 else "VC-WIN64A "

        self.exec_vs(
            r"%(perl_dir)s\bin\perl.exe Configure "
            + debug_option
            + target_option
            + common_options
        )

        with contextlib.suppress(Exception):
            self.exec_vs(r"nmake /nologo clean")
        self.exec_vs(r"nmake /nologo")
        self.exec_vs(r"%(perl_dir)s\bin\perl.exe mk-ca-bundle.pl -n cert.pem")
        self.exec_vs(r"nmake /nologo install")

        self.install(r".\cert.pem bin")
        self.install(r".\LICENSE share\doc\openssl")
        self.install_pc_files()


@project_add
class OpenSSL(Tarball, Project):
    def __init__(self):
        Project.__init__(
            self,
            "openssl",
            version="3.0.8",
            archive_url="https://www.openssl.org/source/old/{major}.{minor}/openssl-{version}.tar.gz",
            hash="6c13d2bf38fdf31eac3ce2a347073673f5d63263398f1f69d0df4a41253e4b3e",
            dependencies=[
                "openssl-base",
            ],
        )

    def build(self):
        common_options = "enable-fips no-ssl3 no-comp --openssldir=%(gtk_dir)s/etc/ssl --prefix=%(gtk_dir)s"
        debug_option = "debug-" if self.builder.opts.configuration == "debug" else ""
        target_option = "VC-WIN32 " if self.builder.x86 else "VC-WIN64A "

        self.exec_vs(
            r"%(perl_dir)s\bin\perl.exe Configure "
            + debug_option
            + target_option
            + common_options
        )

        with contextlib.suppress(Exception):
            self.exec_vs(r"nmake /nologo clean")
        self.exec_vs(r"nmake /nologo")
        self.exec_vs(r"nmake /nologo install_fips")



#  Copyright (C) 2016 The Gvsbuild Authors
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, see <http://www.gnu.org/licenses/>.


"""Main build script."""

# Verify we can import from the script directory
import rich
import typer

from gvsbuild.deps import deps
from gvsbuild.list import list_

try:
    import gvsbuild.utils.utils  # noqa: F401
except ImportError:
    # We are probably using an embedded installation
    print("Error importing utility, fixing paths ...")
    import os
    import sys

    # Get the script dir
    script_dir = os.path.dirname(os.path.realpath(sys.argv[0]))
    # and add it at the beginning, emulating the standard python startup
    sys.path.insert(0, script_dir)

import gvsbuild.groups  # noqa: F401
import gvsbuild.projects  # noqa: F401
import gvsbuild.tools  # noqa: F401
from gvsbuild.build import build
from gvsbuild.info import version_callback
from gvsbuild.outdated import outdated

rich.reconfigure(markup=False)

app = typer.Typer(help="Build GTK for Windows")
app.command(help="")(build)
app.command(help="")(outdated)
app.command(help="", name="list")(list_)
app.command(help="")(deps)


@app.callback()
def common(
    ctx: typer.Context,
    version: bool = typer.Option(
        None, "--version", callback=version_callback, help="Show the app's version"
    ),
):
    pass


def run():
    app()



#  Copyright (C) 2016 The Gvsbuild Authors
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, see <http://www.gnu.org/licenses/>.

import os
import pathlib
import re
import shutil
import stat
import time

from .simple_ui import log


def convert_to_msys(path):
    path = path
    if path[1] != ":":
        raise NotADirectoryError("Path doesn't contain a drive letter like C:")
    return f"/{path[0]}" + path[2:].replace("\\", "/")


def _rmtree_error_handler(func, path, exc_info):
    if not os.access(path, os.W_OK):
        # Is the error an access error ?
        os.chmod(path, stat.S_IWUSR)
        func(path)
        log.debug(f"rmtree:read-only file/path ({path})")
    else:
        raise


def rmtree_full(dest_dir, retry=False):
    if not os.path.isdir(dest_dir):
        return
    if retry:
        for delay in [0.1, 0.2, 0.4, 0.8]:
            try:
                shutil.rmtree(dest_dir, onerror=_rmtree_error_handler)
                break
            except OSError:
                # wait a little, don't ask me why ;(
                time.sleep(delay)
    else:
        shutil.rmtree(dest_dir, onerror=_rmtree_error_handler)


def read_file(file_name):
    with open(file_name, encoding="utf-8") as fi:
        rt = [line.rstrip("\n") for line in fi]
    return rt


def write_file(file_name, content):
    with open(file_name, "w", encoding="utf-8") as fo:
        for i in content:
            fo.write(f"{i}\n")


def file_replace(file_name, chg_list, make_bak=True):
    """Execute a series of replace on the file indicated.

    chg_list is an iterable of tuple (find, replace) to execute
    """

    fc = read_file(file_name)
    if make_bak:
        sv = fc
    chg = 0
    for find, repl in chg_list:
        exp = re.compile(find)
        nw = []
        for i in fc:
            nl = exp.sub(repl, i)
            if nl != i:
                chg += 1
            nw.append(nl)
        fc = nw
    if chg:
        if make_bak:
            write_file(f"{file_name}.bak", sv)
        write_file(file_name, fc)


class ordered_set(set):
    def __init__(self):
        set.__init__(self)
        self.__list = []

    def add(self, o):
        if o not in self:
            set.add(self, o)
            self.__list.append(o)

    def remove(self, o):
        if o in self:
            set.remove(self, o)
            self.__list.remove(o)

    def __iter__(self):
        return self.__list.__iter__()


def python_find_libs_dir(org_dir):
    """From the python org_dir that can be also a virtualenv path return the
    libs dir."""

    cur = os.path.join(org_dir, "libs")
    if os.path.isdir(cur):
        # easy :)
        return cur

    # look for the virtualenv marker
    chk = os.path.join(org_dir, "lib")
    if not os.path.isdir(chk):
        # one level up
        chk = os.path.join(org_dir, "..", "lib")

    if not chk:
        # oops
        return None

    orig_file = os.path.join(chk, "orig-prefix.txt")
    if os.path.isfile(orig_file):
        # Read and see what's happening
        org_dir = pathlib.Path(orig_file).read_text()
    # Let's see if now is ok ..
    cur = os.path.join(org_dir, "libs")
    return cur if os.path.isdir(cur) else None



#  Copyright (C) 2016 The Gvsbuild Authors
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, see <http://www.gnu.org/licenses/>.

from gvsbuild.utils.base_builders import Meson
from gvsbuild.utils.base_expanders import Tarball
from gvsbuild.utils.base_project import Project, project_add


@project_add
class GtkSourceView4(Tarball, Meson):
    def __init__(self):
        Project.__init__(
            self,
            "gtksourceview4",
            version="4.8.4",
            lastversion_major=4,
            lastversion_even=True,
            repository="https://gitlab.gnome.org/GNOME/gtksourceview",
            archive_url="https://download.gnome.org/sources/gtksourceview/{major}.{minor}/gtksourceview-{version}.tar.xz",
            hash="7ec9d18fb283d1f84a3a3eff3b7a72b09a10c9c006597b3fbabbb5958420a87d",
            dependencies=["meson", "ninja", "gtk3", "pkgconf", "libxml2"],
        )
        if Project.opts.enable_gi:
            self.add_dependency("gobject-introspection")
        else:
            self.add_param("-Dgir=false")
        self.add_param("-Dvapi=false")

    def build(self):
        Meson.build(self)
        self.install(r".\COPYING share\doc\gtksourceview4")


@project_add
class GtkSourceView5(Tarball, Meson):
    def __init__(self):
        Project.__init__(
            self,
            "gtksourceview5",
            version="5.12.0",
            lastversion_major=5,
            lastversion_even=True,
            repository="https://gitlab.gnome.org/GNOME/gtksourceview",
            archive_url="https://download.gnome.org/sources/gtksourceview/{major}.{minor}/gtksourceview-{version}.tar.xz",
            hash="daf32ff5d3150d6385917d3503a85b9e047ba158b2b03079314c9c00813fa01f",
            dependencies=["meson", "ninja", "gtk4", "pkgconf", "libxml2"],
            patches=[
                "001-fix-fontconfig-availability-check.patch",
            ],
        )
        if Project.opts.enable_gi:
            self.add_dependency("gobject-introspection")
        else:
            self.add_param("-Dintrospection=disabled")
        self.add_param("-Dvapi=false")

    def build(self):
        Meson.build(self)
        self.install(r".\COPYING share\doc\gtksourceview5")



#  Copyright (C) 2016 The Gvsbuild Authors
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, see <http://www.gnu.org/licenses/>.

import os

from gvsbuild.utils.base_expanders import Tarball
from gvsbuild.utils.base_project import Project, project_add


@project_add
class Icu(Tarball, Project):
    def __init__(self):
        Project.__init__(
            self,
            "icu",
            repository="https://github.com/unicode-org/icu",
            lastversion_major="74",
            version="74.2",
            archive_url="https://github.com/unicode-org/icu/releases/download/release-{major}-{minor}/icu4c-{major}_{minor}-src.zip",
            hash="b22e94977a82aac7ebe269ee00bc2d3164bd4495cafcb9e0b2109ab7fac2a37d",
        )

    def build(self):
        bindir = r".\bin"
        libdir = r".\lib"
        if not self.builder.x86:
            bindir += "64"
            libdir += "64"
        if self.opts.vs_ver != "15":
            # Not Vs2017, we change the platform
            search, replace = self._msbuild_make_search_replace(141)
            self._msbuild_copy_dir(
                None,
                os.path.join(self.build_dir, "source", "allinone"),
                search,
                replace,
            )

        self.exec_msbuild(r"source\allinone\allinone.sln /t:cal /t:MakeData")

        if self.builder.opts.configuration == "debug":
            self.install_pc_files("pc-files-debug")
        else:
            self.install_pc_files()

        self.install(r".\LICENSE share\doc\icu")
        self.install(bindir + r"\* bin")
        self.install(libdir + r"\* lib")
        self.install(r".\include\* include")



#  Copyright (C) 2016 The Gvsbuild Authors
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, see <http://www.gnu.org/licenses/>.
from gvsbuild.utils.base_builders import Meson
from gvsbuild.utils.base_expanders import Tarball
from gvsbuild.utils.base_project import Project, project_add


@project_add
class Libsigcplusplus(Tarball, Meson):
    def __init__(self):
        Project.__init__(
            self,
            "libsigc++",
            prj_dir="libsigc++",
            version="3.6.0",
            lastversion_even=True,
            repository="https://github.com/libsigcplusplus/libsigcplusplus",
            archive_url="https://github.com/libsigcplusplus/libsigcplusplus/releases/download/{version}/libsigc++-{version}.tar.xz",
            hash="c3d23b37dfd6e39f2e09f091b77b1541fbfa17c4f0b6bf5c89baef7229080e17",
            dependencies=[
                "meson",
                "ninja",
            ],
        )

    def build(self):
        Meson.build(
            self,
            meson_params="-Dbuild-examples=false -Dbuild-documentation=false",
        )

        self.install(r".\COPYING share\doc\libsigc++")



#  Copyright (C) 2016 The Gvsbuild Authors
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, see <http://www.gnu.org/licenses/>.

from gvsbuild.utils.base_builders import Meson
from gvsbuild.utils.base_expanders import Tarball
from gvsbuild.utils.base_project import Project, project_add


@project_add
class Libepoxy(Tarball, Meson):
    def __init__(self):
        Project.__init__(
            self,
            "libepoxy",
            version="1.5.10",
            archive_url="https://github.com/anholt/libepoxy/archive/refs/tags/{version}.tar.gz",
            archive_filename="libepoxy-{version}.tar.gz",
            hash="a7ced37f4102b745ac86d6a70a9da399cc139ff168ba6b8002b4d8d43c900c15",
            dependencies=["ninja", "meson"],
        )

    def build(self):
        Meson.build(self)
        self.install(r"COPYING share\doc\libepoxy")



#  Copyright (C) 2016 The Gvsbuild Authors
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, see <http://www.gnu.org/licenses/>.

"""Base project class, used also for tools."""

import datetime
import os
import pathlib
import re
import shutil
from enum import Enum
from typing import Dict, Generic, List, Tuple, TypeVar

from .simple_ui import log
from .utils import _rmtree_error_handler


class ProjectType(str, Enum):
    IGNORE = "ignore"
    PROJECT = "project"
    TOOL = "tool"
    GROUP = "group"


class Options:
    def __init__(self):
        self.enable_gi = False
        self.ffmpeg_enable_gpl = False
        self.verbose = False
        self.debug = False
        self.platform = "x64"
        self.configuration = "release"
        self.release_configuration_is_actually_debug_optimized = False
        self.build_dir = None
        self.archives_download_dir = None
        self.export_dir = None
        self.patches_root_dir = None
        self.tools_root_dir = None
        self.vs_ver = None
        self.vs_install_path = None
        self.win_sdk_ver = None
        self.net_target_framework = None
        self.net_target_framework_version = None
        self.msys_dir = None
        self.clean = False
        self.msbuild_opts = None
        self.use_env = False
        self.deps = False
        self.check_hash = False
        self.skip = False
        self.make_zip = False
        self.zip_continue = False
        self.from_scratch = False
        self.keep_tools = False
        self.fast_build = False
        self.keep_going = False
        self.clean_built = False
        self.py_wheel = False
        self.log_size = None
        self.log_single = False
        self.cargo_opts = None
        self.ninja_opts = None
        self.capture_out = False
        self.print_out = False
        self.git_expand_dir = None
        self.projects = None


P = TypeVar("P")


class Project(Generic[P]):
    def __init__(self, name, **kwargs):
        object.__init__(self)
        self.patch_dir = None
        self.build_dir = None
        self.pkg_dir = None
        self.builder = None
        self.name = name
        self.prj_dir = name
        self.dependencies = []
        self.patches = []
        self.archive_url = None
        self.archive_filename = None
        self.tarbomb = False
        self.type = None
        self.version = None
        self.repository = None
        self.lastversion_major = None
        self.lastversion_even = None
        self.internal = False
        self.mark_file = None
        self.clean = False
        self.to_add = True
        self.extra_env = {}
        self.tag = None
        self.repo_url = None
        self.archive_filename = None

        for k in kwargs:
            setattr(self, k, kwargs[k])
        self.__working_dir = None
        if len(self.name) > Project.name_len:
            Project.name_len = len(self.name)

        if not self.version:
            self.version = f"git/{self.tag}" if self.repo_url else "undefined"
        version_params = {
            "version": self.version,
            "tag": self.tag,
        }
        match = re.match(
            r"(?P<major>\d+)(\.(?P<minor>\d+))?(\.(?P<micro>\d+))?", self.version
        )
        if match:
            for param in ["major", "minor", "micro"]:
                version_params[param] = match[param] or ""

        if self.archive_url:
            self.archive_url = self.archive_url.format(**version_params)
        if self.archive_filename:
            self.archive_filename = self.archive_filename.format(**version_params)

        # register version params for use from derived classes
        self.version_params = version_params

    _projects: List[P] = []
    _names: List[str] = []
    _dict: Dict[str, P] = {}
    _ver_res = None
    name_len = 0
    # List of class/type to add, now not at import time but after some options are parsed
    _reg_prj_list: List[Tuple[P, ProjectType]] = []
    # build option
    opts = Options()

    @staticmethod
    def compute_dependencies(projects):
        global_deps = {p.name for p in projects}

        def _add_project_dependencies(project):
            for dep in project.dependencies:
                if dep not in global_deps:
                    global_deps.add(dep)
                    _add_project_dependencies(Project.get_project(dep))

        for project in projects:
            _add_project_dependencies(project)

        return [Project.get_project(p) for p in global_deps]

    def __str__(self):
        return self.name

    def __repr__(self):
        return repr(self.name)

    def load_defaults(self):
        # Used by the tools to load default paths/filenames
        pass

    def finalize_dep(self, builder, deps):
        """Used to manipulate the dependencies list, to add or remove projects
        For the dev-shell project is used to limit the tools to use."""
        pass

    def build(self):
        raise NotImplementedError("build")

    def post_install(self):
        pass

    def add_dependency(self, dep):
        self.dependencies.append(dep)

    def exec_cmd(self, cmd, working_dir=None, add_path=None):
        self.builder.exec_cmd(cmd, working_dir=working_dir, add_path=add_path)

    def exec_vs(self, cmd, add_path=None):
        self.builder.exec_vs(
            cmd, working_dir=self._get_working_dir(), add_path=add_path
        )

    def exec_msbuild(self, cmd, configuration=None, add_path=None):
        if not configuration:
            configuration = "%(configuration)s"
        self.exec_vs(
            "msbuild "
            + cmd
            + " /p:Configuration="
            + configuration
            + " %(msbuild_opts)s",
            add_path=add_path,
        )

    def _msbuild_make_search_replace(self, org_platform):
        """Return the search & replace strings (converted to bytes to update
        the platform Toolset version (v140, v141, ...) to use a new compiler,
        e.g. to use vs2017 solution's files for vs2019.

        The '<PlatformToolset' at the beginning is missing to handle
        projects like libmictohttpd that has a condition in the platform
        definition
        """

        ver = self.builder.opts.vs_ver
        if ver == "15":
            dst_platform = "141"
        elif ver == "16":
            dst_platform = "142"
        elif ver == "17":
            dst_platform = "143"
        else:
            dst_platform = f"{ver}0"
        search = (">v%u</PlatformToolset>" % (org_platform,)).encode("utf-8")
        replace = f">v{dst_platform}</PlatformToolset>".encode()

        return search, replace

    def _msbuild_copy_dir(self, dst, src, search, replace):
        """Converts & copy a directory of a vs solution to be use with a new
        platform toolset & visual studio version.

        If dst is None the change is made in place
        """

        if dst:
            os.makedirs(dst, exist_ok=True)
            copy = True
        else:
            dst = src
            copy = False

        for cf in os.scandir(src):
            src_full = os.path.join(src, cf.name)
            dst_full = os.path.join(dst, cf.name)

            if cf.is_file():
                with open(src_full, "rb") as f:
                    content = f.read()
                new_content = content.replace(search, replace)
                if content != new_content:
                    log.info(f"File changed ({src_full})")
                    write = True
                else:
                    log.info(f"   same file ({src_full})")
                    write = copy

                if write:
                    dst_full = os.path.join(dst, cf.name)
                    with open(dst_full, "wb") as f:
                        f.write(new_content)
            elif cf.is_dir():
                self._msbuild_copy_dir(
                    dst_full if copy else None, src_full, search, replace
                )

    def exec_msbuild_gen(
        self,
        base_dir,
        sln_file,
        add_pars="",
        configuration=None,
        add_path=None,
        use_env=False,
    ):
        r"""looks for base_dir\{vs_ver}\sln_file or
        base_dir\{vs_ver_tear}\sln_file for launching the msbuild commamd.

        If it's not present in the directory the system start to look
        backward to find the first version present
        """

        def _msbuild_ok(self, dir_part):
            full = os.path.join(self.build_dir, base_dir, dir_part, sln_file)
            log.info(f"Checking for '{full}'")
            return os.path.exists(full)

        def _msbuild_copy(self, org_path, org_platform, use_ver=True):
            if use_ver:
                dst_part = f"vs{self.builder.opts.vs_ver}"
            else:
                dst_part = self.builder.vs_ver_year
            dst = os.path.join(self.build_dir, base_dir, dst_part)
            src = os.path.join(self.build_dir, base_dir, org_path)
            search, replace = self._msbuild_make_search_replace(org_platform)
            log.info(f"Vs solution copy: '{src}' -> '{dst}'")
            self._msbuild_copy_dir(dst, src, search, replace)
            return dst_part

        part = f"vs{self.builder.opts.vs_ver}"
        if not _msbuild_ok(self, part):
            part = self.builder.vs_ver_year
            if not _msbuild_ok(self, part):
                part = None

        if not part:
            look = {
                "12": [],
                "14": [
                    (
                        "vs12",
                        120,
                        True,
                    ),
                    (
                        "vs2013",
                        120,
                        False,
                    ),
                ],
                "15": [
                    (
                        "vs14",
                        140,
                        True,
                    ),
                    (
                        "vs2015",
                        140,
                        False,
                    ),
                    (
                        "vs12",
                        120,
                        True,
                    ),
                    (
                        "vs2013",
                        120,
                        False,
                    ),
                ],
                "16": [
                    (
                        "vs15",
                        141,
                        True,
                    ),
                    (
                        "vs2017",
                        141,
                        False,
                    ),
                    (
                        "vs14",
                        140,
                        True,
                    ),
                    (
                        "vs2015",
                        140,
                        False,
                    ),
                    (
                        "vs12",
                        120,
                        True,
                    ),
                    (
                        "vs2013",
                        120,
                        False,
                    ),
                ],
            }
            lst = look.get(self.builder.opts.vs_ver, [])
            for p in lst:
                if _msbuild_ok(self, p[0]):
                    # Found one, create the new directory with a copy, changing the platform identifier
                    part = _msbuild_copy(self, p[0], p[1], p[2])
                    break
            if part:
                # We log what we found because is not the default
                log.log(f"Project {self.name}, using {part} directory")

        if part:
            cmd = os.path.join(base_dir, part, sln_file)
            if add_pars:
                cmd += f" {add_pars}"
            if use_env:
                cmd += " /p:UseEnv=True"
        else:
            log.error_exit(
                f"Solution file '{sln_file}' for project '{self.name}' not found!"
            )
        self.exec_msbuild(cmd, configuration, add_path)
        return part

    def install(self, *args):
        self.builder.install(self._get_working_dir(), self.pkg_dir, *args)

    def install_dir(self, src, dest=None):
        if not dest:
            dest = os.path.basename(src)
        self.builder.install_dir(self._get_working_dir(), self.pkg_dir, src, dest)

    def install_pc_files(self, base_dir="pc-files"):
        """Install, setting dir & version, the .pc files."""
        pkgconfig_dir = os.path.join(self.pkg_dir, "lib", "pkgconfig")
        self.builder.make_dir(pkgconfig_dir)

        src_dir = os.path.join(self._get_working_dir(), base_dir)
        log.debug(f"Copy .pc files from {src_dir} to {pkgconfig_dir}")
        gtk_dir = self.builder.gtk_dir.replace("\\", "/")
        for f in os.scandir(src_dir):
            if f.is_file():
                log.debug(f" {f.name}")
                content = pathlib.Path(f.path).read_text()
                _t = content.replace("@prefix@", gtk_dir)
                content = _t
                _t = content.replace("@version@", self.version)
                content = _t

                with open(
                    os.path.join(pkgconfig_dir, f.name), "w", encoding="utf-8"
                ) as fo:
                    fo.write(content)

    def patch(self):
        for p in self.patches:
            name = os.path.basename(p)
            stamp = os.path.join(self.build_dir, f"{name}.patch-applied")
            if not os.path.exists(stamp):
                log.log(f"Applying patch {p}")
                self.builder.exec_msys(
                    ["patch", "-p1", "-i", p], working_dir=self._get_working_dir()
                )
                with open(stamp, "w", encoding="utf-8") as stampfile:
                    stampfile.write("done")
            else:
                log.debug(f"patch {p} already applied, skipping")

    def _get_working_dir(self):
        if self.__working_dir:
            return os.path.join(self.build_dir, self.__working_dir)
        else:
            return self.build_dir

    def push_location(self, path):
        self.__working_dir = path

    def pop_location(self):
        self.__working_dir = None

    def prepare_build_dir(self):
        if self.clean and os.path.exists(self.build_dir):
            shutil.rmtree(self.build_dir, onerror=_rmtree_error_handler)

        if os.path.exists(self.build_dir):
            log.debug(f"directory {self.build_dir} already exists")
            if self.update_build_dir():
                self.mark_file_remove()
                if os.path.exists(self.patch_dir):
                    log.log(f"Copying files from {self.patch_dir} to {self.build_dir}")
                    self.builder.copy_all(self.patch_dir, self.build_dir)
        else:
            self.unpack()
            if os.path.exists(self.patch_dir):
                log.log(f"Copying files from {self.patch_dir} to {self.build_dir}")
                self.builder.copy_all(self.patch_dir, self.build_dir)

    def update_build_dir(self):
        pass

    def unpack(self):
        raise NotImplementedError("unpack")

    def export(self):
        raise NotImplementedError("export")

    def get_path(self):
        # Optional for projects
        pass

    def add_extra_env(self, key, val):
        # Extra env vars for projects / tools
        self.extra_env[key] = val

    def apply_extra_env(self, base_env):
        if self.extra_env:
            for key, val in self.extra_env.items():
                if key not in base_env:
                    base_env[key] = val

    @staticmethod
    def add(proj, type=ProjectType.IGNORE):
        Project._projects.append(proj)
        Project._names.append(proj.name)
        Project._dict[proj.name] = proj
        if proj.type is None:
            proj.type = type

    @staticmethod
    def register(cls, ty):
        """Register the class to be added after some initialization."""
        Project._reg_prj_list.append(
            (
                cls,
                ty,
            )
        )

    @staticmethod
    def add_all():
        """Add all the registered class."""
        for cls, ty in Project._reg_prj_list:
            c_inst = cls()
            if c_inst.to_add:
                Project.add(c_inst, type=ty)
            else:
                del c_inst

    def ignore(self):
        """Mark the project not to build/add to the list."""
        self.to_add = False

    @staticmethod
    def get_project(name):
        try:
            return Project._dict[name]
        except KeyError:
            log.error_exit(f"Could not find project {name}")

    @staticmethod
    def list_projects():
        return list(Project._projects)

    @staticmethod
    def get_names():
        return list(Project._names)

    @staticmethod
    def get_dict():
        return dict(Project._dict)

    @staticmethod
    def get_tool_path(tool):
        if not isinstance(tool, Project):
            tool = Project._dict[tool]
        if tool.type != ProjectType.TOOL:
            return None
        t = tool.get_path()
        return t[0] or t[1] if isinstance(t, tuple) else t

    @staticmethod
    def get_tool_executable(tool):
        if not isinstance(tool, Project):
            tool = Project._dict[tool]

        return tool.get_executable() if tool.type == ProjectType.TOOL else None

    @staticmethod
    def get_tool_base_dir(tool):
        if not isinstance(tool, Project):
            tool = Project._dict[tool]

        return tool.get_base_dir() if tool.type == ProjectType.TOOL else None

    def mark_file_calc(self):
        if not self.mark_file:
            self.mark_file = os.path.join(self.build_dir, ".wingtk-built")

    def mark_file_remove(self):
        self.mark_file_calc()
        if os.path.isfile(self.mark_file):
            os.remove(self.mark_file)

    def mark_file_write(self):
        self.mark_file_calc()
        try:
            with open(self.mark_file, "w", encoding="utf-8") as fo:
                now = datetime.datetime.now().replace(microsecond=0)
                fo.write(f"{now.strftime('%Y-%m-%d %H:%M:%S')}\n")
        except FileNotFoundError as e:
            log.debug(f"Exception writing file '{self.mark_file}' ({e})")

    def mark_file_exist(self):
        rt = None
        self.mark_file_calc()
        if os.path.isfile(self.mark_file):
            try:
                with open(self.mark_file, encoding="utf-8") as fi:
                    rt = fi.readline().strip("\n")
            except OSError as e:
                print(f"Exception reading file '{self.mark_file}'")
                print(e)
        return rt

    def is_project(self):
        return self.type == ProjectType.PROJECT


def project_add(cls):
    """Class decorator to add the newly created Project class to the global
    projects/tools/groups list."""
    Project.register(cls, ProjectType.PROJECT)
    return cls


def get_project_by_type(prj_type):
    return [
        (project.name, project.version)
        for project in Project._projects
        if project.type == prj_type
    ]



#  Copyright (C) 2016 The Gvsbuild Authors
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, see <http://www.gnu.org/licenses/>.

from gvsbuild.utils.base_expanders import Tarball
from gvsbuild.utils.base_project import Project, project_add


@project_add
class SQLite(Tarball, Project):
    def __init__(self):
        Project.__init__(
            self,
            "sqlite",
            version="3.45.2",
            archive_url="https://www.sqlite.org/2024/sqlite-autoconf-{major}{minor:0<3}{micro:0<3}.tar.gz",
            hash="bc9067442eedf3dd39989b5c5cfbfff37ae66cc9c99274e0c3052dc4d4a8f6ae",
        )

    def build(self):
        nmake_debug = (
            "DEBUG=2" if self.builder.opts.configuration == "debug" else "DEBUG=0"
        )
        self.exec_vs(f"nmake /f Makefile.msc sqlite3.dll DYNAMIC_SHELL=1 {nmake_debug}")

        self.install("sqlite3.h include")
        self.install("sqlite3ext.h include")
        self.install("sqlite3.dll sqlite3.pdb bin")
        self.install("sqlite3.lib lib")



#  Copyright (C) 2016 The Gvsbuild Authors
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, see <http://www.gnu.org/licenses/>.

from gvsbuild.utils.base_expanders import Tarball
from gvsbuild.utils.base_project import Project, project_add


@project_add
class Lz4(Tarball, Project):
    def __init__(self):
        Project.__init__(
            self,
            "lz4",
            version="1.9.4",
            archive_url="https://github.com/lz4/lz4/archive/v{version}.tar.gz",
            archive_filename="lz4-{version}.tar.gz",
            hash="0b0e3aa07c8c063ddf40b082bdf7e37a1562bda40a0ff5272957f3e987e0e54b",
        )

    def build(self):
        self.exec_msbuild_gen(r"visual", "lz4.sln")

        self.install(
            r"visual\%(vs_ver_year)s\bin\%(platform)s_%(configuration)s\liblz4.dll visual\%(vs_ver_year)s\bin\%(platform)s_%(configuration)s\liblz4.pdb bin"
        )
        self.install(r".\lib\lz4.h .\lib\lz4hc.h .\lib\lz4frame.h include")
        self.install(
            r"visual\%(vs_ver_year)s\bin\%(platform)s_%(configuration)s\liblz4.lib lib"
        )

        self.install(r".\lib\LICENSE share\doc\lz4")
        self.install_pc_files()



#  Copyright (C) 2016 The Gvsbuild Authors
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, see <http://www.gnu.org/licenses/>.

import os

from gvsbuild.utils.base_expanders import Tarball
from gvsbuild.utils.base_project import Project, project_add
from gvsbuild.utils.utils import convert_to_msys


@project_add
class Libvpx(Tarball, Project):
    def __init__(self):
        Project.__init__(
            self,
            "libvpx",
            version="1.14.0",
            archive_url="https://github.com/webmproject/libvpx/archive/v{version}.tar.gz",
            archive_filename="libvpx-v{version}.tar.gz",
            hash="5f21d2db27071c8a46f1725928a10227ae45c5cd1cad3727e4aafbe476e321fa",
            dependencies=["nasm", "msys2", "libyuv", "perl"],
            patches=[
                "0006-gen_msvs_vcxproj.sh-Select-current-Windows-SDK-if-av.patch",
                "0001-Always-generate-pc-file.patch",
            ],
        )

    def build(self):
        configure_options = (
            "--enable-pic --as=nasm --disable-unit-tests --size-limit=16384x16384 "
            "--enable-postproc --enable-multi-res-encoding --enable-temporal-denoising "
            "--enable-vp9-temporal-denoising --enable-vp9-postproc --disable-tools "
            "--disable-examples --disable-docs "
        )
        if self.builder.opts.configuration == "debug":
            configure_options += "--enable-debug_libs"

        target = "x86-win32-vs" if self.builder.x86 else "x86_64-win64-vs"
        target += self.builder.opts.vs_ver

        msys_path = Project.get_tool_path("msys2")

        self.exec_vs(
            r"%s\bash ./configure --target=%s --prefix=%s %s"
            % (
                msys_path,
                target,
                convert_to_msys(self.builder.gtk_dir),
                configure_options,
            ),
            add_path=msys_path,
        )
        self.exec_vs(r"make", add_path=msys_path)
        self.exec_vs(r"make install", add_path=msys_path)

        self.install(r".\LICENSE share\doc\libvpx")

    def post_install(self):
        # LibVPX generates a static library named 'vpxmd.lib' or 'vpxmdd.lib'
        # in an unusual directory which is not the same as expected by the vpx.pc file
        if self.builder.opts.configuration == "debug":
            lib_name = "vpxmdd.lib"
        else:
            lib_name = "vpxmd.lib"
        lib_path = f"Win32/{lib_name}" if self.builder.x86 else f"x64/{lib_name}"
        self.builder.exec_msys(
            ["mv", lib_path, "./vpx.lib"],
            working_dir=os.path.join(self.builder.gtk_dir, "lib"),
        )



#  Copyright (C) 2016 The Gvsbuild Authors
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, see <http://www.gnu.org/licenses/>.

"""Main builder class."""

import contextlib
import copy
import glob
import hashlib
import json
import os
import re
import shutil
import ssl
import subprocess
import sys
import time
import traceback
from pathlib import Path
from typing import Optional
from urllib.error import ContentTooShortError, URLError
from urllib.request import urlopen

from .base_expanders import dirlist2set, make_zip
from .base_project import Project
from .simple_ui import log, script_title
from .utils import ordered_set, rmtree_full


class Builder:
    def __init__(self, opts):
        self.opts = opts

        script_title("* Setup")

        # Check and normalize the platform
        if opts.platform in ("Win32", "win32", "x86"):
            opts.platform = "Win32"
            opts.x86 = True
            self.filename_arch = "x86"
        elif opts.platform in ("x64", "amd64", "Amd64"):
            opts.platform = "x64"
            self.filename_arch = "x64"
            opts.x86 = False
        else:
            raise NameError(f"Invalid target platform '{opts.platform}'")

        opts.x64 = not opts.x86
        # Setup the directory, used by check vs
        self.working_dir = os.path.join(
            opts.build_dir, "build", opts.platform, opts.configuration
        )
        self.gtk_dir = os.path.join(
            opts.build_dir, "gtk", opts.platform, opts.configuration
        )

        if opts.from_scratch:
            with log.simple_oper("Cleanup build directories"):
                with log.simple_oper(
                    f"Removing working/building dir ({self.working_dir})"
                ):
                    rmtree_full(self.working_dir, retry=True)
                with log.simple_oper(f"Removing destination dir ({self.gtk_dir})"):
                    rmtree_full(self.gtk_dir, retry=True)
                with log.simple_oper(
                    f"Removing git expand dir ({self.opts.git_expand_dir})"
                ):
                    rmtree_full(self.opts.git_expand_dir, retry=True)
                with log.simple_oper(
                    f"Removing archives download dir ({opts.archives_download_dir})"
                ):
                    rmtree_full(opts.archives_download_dir)
                if not opts.keep_tools:
                    with log.simple_oper(f"Removing tools dir ({opts.tools_root_dir})"):
                        rmtree_full(opts.tools_root_dir, retry=True)
                else:
                    log.message(f"Keeping tools dir ({opts.tools_root_dir})")

        if not opts.use_env:
            self.__minimum_env()

        self.x86 = opts.platform == "Win32"
        self.x64 = not self.x86

        # Create the year version for Visual Studio
        vs_zip_parts = {
            "12": "vs2013",
            "14": "vs2015",
            "15": "vs2017",
            "16": "vs2019",
            "17": "vs2022",
        }

        self.vs_ver_year = vs_zip_parts.get(opts.vs_ver)
        if not self.vs_ver_year:
            self.vs_ver_year = f"ms-cl-{opts.vs_ver}"

        vs_part = self.vs_ver_year
        if opts.win_sdk_ver:
            vs_part += f"-{opts.win_sdk_ver}"

        self.zip_dir = os.path.join(
            opts.build_dir, "dist", vs_part, opts.platform, opts.configuration
        )
        if opts.make_zip:
            if opts.zip_continue:
                self.file_built = self._load_built_files()
            else:
                # Remove the destination dir before starting anything
                if os.path.isdir(self.gtk_dir):
                    with log.simple_oper(f"Removing build dir ({self.gtk_dir})"):
                        rmtree_full(self.gtk_dir, retry=True)
                self.file_built = set()
            os.makedirs(self.zip_dir, exist_ok=True)

        self.__check_tools(opts)
        self.__check_vs(opts)

    def _create_msbuild_opts(self, python):
        rt = [f"/nologo /p:Platform={self.opts.platform}"]
        if python:
            rt.append(
                '/p:PythonPath="%(python_dir)s" /p:PythonDir="%(python_dir)s"'
                % {"python_dir": python}
            )

        if log.verbose_on():
            rt.append("/v:normal")
        else:
            rt.append("/v:minimal")

        if self.opts.win_sdk_ver:
            rt.append(f'/p:WindowsTargetPlatformVersion="{self.opts.win_sdk_ver}"')

        if self.opts.net_target_framework:
            rt.append(f'/p:TargetFrameworks="{self.opts.net_target_framework}"')

        if self.opts.net_target_framework_version:
            rt.append(
                f'/p:TargetFrameworkVersion="{self.opts.net_target_framework_version}"'
            )

        if self.opts.msbuild_opts:
            rt.append(self.opts.msbuild_opts)

        return " ".join(rt)

    def __minimum_env(self):
        r"""Set the environment to the minimum needed to run, leaving only the
        c:\windows\XXXX directory and the git one.

        The LIB, LIBPATH & INCLUDE environment are also cleaned to avoid
        mismatch with  libs / programs already installed
        """

        log.start("Cleaning up the build environment")
        win_dir = os.environ.get("SYSTEMROOT", r"c:\windows").lower()

        win_dir = win_dir.replace("\\", "\\\\")
        log.debug(f"windir -> {win_dir}")

        chk_re = [
            re.compile(f"^{win_dir}\\\\"),
            re.compile(f"^{win_dir}$"),
            re.compile("\\\\git\\\\"),
            re.compile("\\\\git$"),
        ]

        mp = []
        paths = os.environ.get("PATH", "").split(";")
        for k in paths:
            # use all lower
            k = os.path.normpath(k).lower()
            if k in mp:
                # already present
                log.debug(f"   Already present: '{k}'")
                continue

            add = False
            for cre in chk_re:
                if cre.search(k):
                    mp.append(k)
                    add = True
                    break
            if add:
                log.debug(f"Add '{k}'")
            else:
                log.debug(f"   Skip '{k}'")

        log.debug("Final path:")
        for i in mp:
            log.debug(f"    {i}")
        os.environ["PATH"] = ";".join(mp)

        os.environ["LIB"] = ""
        os.environ["LIBPATH"] = ""
        os.environ["INCLUDE"] = ""
        log.end()

    def __msys_missing(self, base_dir: Path) -> Optional[list]:
        msys_pkg = [
            ("patch", "patch"),
            ("make", "make"),
            ("md5sum", "coreutils"),
            ("diff", "diffutils"),
            ("bison", "bison"),
            ("flex", "flex"),
        ]
        missing = []
        for prog, pkg in msys_pkg:
            if not Path.is_file(base_dir / "usr" / "bin" / f"{prog}.exe"):
                log.log(f"msys: missing package {pkg}")
                missing.append(pkg)
        return missing

    def __check_tools(self, opts):
        script_title("* Msys tool")
        log.start("Checking msys tool")
        msys_path = opts.msys_dir
        if not msys_path or not Path.exists(msys_path):
            msys_paths = [
                Path(r"C:\msys64"),
                Path(r"C:\msys32"),
                Path(r"C:\tools\msys64"),
                Path(r"C:\tools\msys32"),
            ]
            for path in msys_paths:
                if Path.exists(path):
                    msys_path = path
                    self.opts.msys_dir = str(msys_path)
                    break
        log.message(f"Using {msys_path} for msys installation")
        # what's missing ?
        missing = self.__msys_missing(msys_path)
        if missing:
            # install using pacman
            cmd = (
                str(msys_path / "usr" / "bin" / "bash")
                + ' -l -c "pacman --noconfirm -S '
                + " ".join(missing)
                + '"'
            )
            log.debug(f"Updating msys2 with '{cmd}'")
            subprocess.check_call(cmd, shell=True)
            missing = self.__msys_missing(msys_path)
        if missing:
            # oops
            cmd = "pacman -S " + " ".join(missing)
            log.error_exit(
                "Missing package(s) from msys2 installation, try with\n    '%s'\nin a msys2 shell."
                % (cmd,)
            )

        self.patch = msys_path / "usr" / "bin" / "patch.exe"
        if not Path.exists(self.patch):
            log.error_exit(
                f"{str(self.patch)} not found. Please check that you installed patch in msys2 using ``pacman -S patch``"
            )

        log.debug(f"patch: {self.patch}")
        log.end()

    def _add_env(self, key, value, env, prepend=True, subst=False):
        # env manipulation helper fun
        # returns a tuple with the old (key, value, ) to let the script restore it if needed
        org_env = env.get(key, None)
        te = None if subst else org_env
        if te:
            env[key] = f"{value};{te}" if prepend else f"{te};{value}"
        else:
            # not set or forced
            env[key] = value
        return (
            key,
            org_env,
        )

    def add_global_env(self, key, value, prepend=True):
        # Env to load before the setup for the visual studio environment
        self._add_env(key, value, os.environ, prepend)

    def add_gtk_dir(self, value):
        return f"{self.gtk_dir}\\{value}"

    def mod_env(self, key, value, prepend=True, subst=False, add_gtk=False):
        # Modify the current build environment
        # returns the old value
        if add_gtk:
            value = self.add_gtk_dir(value)
        return self._add_env(key, value, self.vs_env, prepend=prepend, subst=subst)

    def restore_env(self, saved):
        if saved:
            key, value = saved
            if key:
                if value:
                    self.vs_env[key] = value
                else:
                    del self.vs_env[key]

    def __dump_vs_loc(self):
        vswhere = r"%s\Microsoft Visual Studio\Installer\vswhere.exe" % (
            os.environ.get("ProgramFiles(x86)", r"C:\Program Files (x86)")
        )
        log.log("Trying to find Visual Studio installations ...")
        if not os.path.exists(vswhere):
            log.log(f"Could not find vswhere executable ({vswhere})")
            return

        completed_process = subprocess.run(
            [f"{vswhere}", "-all", "-products", "*", "-format", "json", "-utf8"],
            capture_output=True,
            encoding="utf-8",
        )
        try:
            completed_process.check_returncode()
            vs_installs = json.loads(completed_process.stdout)
            return self.__extract_paths(vs_installs)
        except subprocess.CalledProcessError as e:
            log.log(f"Unable to call vswhere.exe to find Visual Studio with error {e}")

    def __extract_paths(self, res):
        log.message("")
        log.message("Visual Studio installation(s) found:")
        paths = []
        for i in res:
            disp = i.get("displayName", "?")
            path = i.get("installationPath", r"?:\?")
            log.message(f"    {disp} @ {path}")
            paths.append(path)
        log.message("")
        return paths

    def __check_good_vs_install(self, opts, vs_path, exit_missing=True):
        # Verify VS exists at the indicated location, and that it supports the required target
        if opts.platform == "Win32":
            vcvars_bat = os.path.join(vs_path, "VC", "bin", "vcvars32.bat")
            # make sure it works with VS 2017+
            if not os.path.exists(vcvars_bat):
                vcvars_bat = os.path.join(
                    vs_path, "VC", "Auxiliary", "Build", "vcvars32.bat"
                )
        else:
            vcvars_bat = os.path.join(vs_path, "VC", "bin", "amd64", "vcvars64.bat")
            # make sure it works with VS Express
            if not os.path.exists(vcvars_bat):
                vcvars_bat = os.path.join(
                    vs_path, "VC", "bin", "x86_amd64", "vcvarsx86_amd64.bat"
                )
            # make sure it works with VS 2017+
            if not os.path.exists(vcvars_bat):
                vcvars_bat = os.path.join(
                    vs_path, "VC", "Auxiliary", "Build", "vcvars64.bat"
                )

        add_opts = f" {opts.win_sdk_ver}" if opts.win_sdk_ver else ""
        log.log(f'Running script "{vcvars_bat}"{add_opts}')
        if not os.path.exists(vcvars_bat):
            if not exit_missing:
                return None

            self.__dump_vs_loc()
            log.error_exit(
                f"\n  {vcvars_bat} could not be found.\n  Please check you have Visual Studio installed at {vs_path}\n  and that it supports the target platform {opts.platform}."
            )
        return subprocess.check_output(
            f'cmd.exe /c ""{vcvars_bat}"{add_opts}>NUL && set"', text=True
        )

    def __find_vs_path_with_vs_version(self, paths):
        for path in paths:
            if self.vs_ver_year[-4:] in path or self.opts.vs_ver in path:
                return path
        log.debug(f"Can't find vs-ver {self.opts.vs_ver} in found VS installations.")

    def __check_vs(self, opts):
        script_title("* Msvc tool")
        log.start("Checking Msvc tool")

        # Add to the environment the GTK paths so meson can find everything
        self.add_global_env("INCLUDE", os.path.join(self.gtk_dir, "include"))
        self.add_global_env("LIB", os.path.join(self.gtk_dir, "lib"))
        self.add_global_env("LIBPATH", os.path.join(self.gtk_dir, "lib"))
        self.add_global_env("PATH", os.path.join(self.gtk_dir, "bin"))

        if not opts.vs_install_path:
            opts.vs_install_path = self.__find_vs_path_with_vs_version(
                self.__dump_vs_loc()
            )
        if opts.vs_install_path is None:
            log.error_exit(
                "Unable to find Visual Studio, try using --vs-ver or --vs-install-path "
                "to specify Visual Studio version or install location"
            )
        log.message(f"Using Visual Studio at {opts.vs_install_path}")
        output = self.__check_good_vs_install(opts, opts.vs_install_path, True)

        self.vs_env = {}
        dbg = log.debug_on()
        for line in output.splitlines():
            e = line.split("=", 1)
            if len(e) < 2:
                log.debug(f"vs env: ignoring {line}")
                continue
            k, v = e
            # Be sure to have PATH in upper case because we need to manipulate it
            if k.upper() == "PATH":
                k = "PATH"
            self.vs_env[k] = v
            if dbg:
                vl = v.split(";")
                if len(vl) > 1:
                    log.debug(f"vs env: {k}: [")
                    for i in vl:
                        log.message_indent(f"  {i}")
                    log.message_indent("]")
                else:
                    log.debug(f"vs env:{k} -> [{v}]")

        if not opts.win_sdk_ver:
            # Try lo figure the sdk version to pass it to the msbuild programs
            sdk = self.vs_env.get("WindowsSDKVersion", "")
            if not sdk or sdk == "\\":
                sdk = self.vs_env.get("WindowsSDKLibVersion", "")
            if sdk:
                # drop the last '\'
                if sdk[-1] == "\\":
                    sdk = sdk[:-1]
                if sdk == "winv6.3":
                    sdk = "8.1"
                opts.win_sdk_ver = sdk
                log.log(f"Auto load win_sdk_ver: '{sdk}'")

        log.end()

    def preprocess(self):
        for proj in Project.list_projects():
            if proj.archive_url:
                if proj.archive_filename:
                    archive = proj.archive_filename
                else:
                    url = proj.archive_url
                    archive = url[url.rfind("/") + 1 :]
                proj.archive_file = os.path.join(
                    self.opts.archives_download_dir, archive
                )
            else:
                proj.archive_file = None
            proj.patch_dir = os.path.join(self.opts.patches_root_dir, proj.prj_dir)
            proj.build_dir = os.path.join(self.working_dir, proj.prj_dir)
            proj.export_dir = self.opts.export_dir
            proj.dependencies = [Project.get_project(dep) for dep in proj.dependencies]
            proj.dependents = []
            proj.load_defaults()
            proj.mark_file_calc()
            if self.opts.clean:
                proj.clean = True

        for proj in Project.list_projects():
            self.__compute_deps(proj)
            log.debug(f"{proj.name} => {[d.name for d in proj.all_dependencies]}")

    def __compute_deps(self, proj):
        if hasattr(proj, "all_dependencies"):
            return
        deps = ordered_set()
        for dep in proj.dependencies:
            self.__compute_deps(dep)
            for p in dep.all_dependencies:
                deps.add(p)
            deps.add(dep)
        proj.finalize_dep(self, deps)
        proj.all_dependencies = deps

    def _drop_proj(self, drop_prj):
        """Delete drop_prj and all the ones that depends on this from the list
        of projects to build."""
        drop_list = [
            drop_prj,
        ]
        while drop_list:
            drop_prj = drop_list.pop(0)

            print(f"* Removing {drop_prj.name} dependents ...")
            for p in self.projects_to_do:
                if drop_prj in p.all_dependencies:
                    print(f"* > Removing {p.name} for {drop_prj.name} ...")
                    # Recursive drop
                    drop_list.append(p)
                    self.projects_to_do.remove(p)
                    self.prj_dropped.append(p.name)

    def build(self, projects):
        if self.__prepare_build(projects):
            return

        if self.opts.check_hash:
            return

        # List of all the project we can mark for build because of a dependend
        self.prj_to_mark = [x for x in Project._projects if x.is_project()]

        self.prj_done = []
        self.prj_skipped = []
        self.prj_err = []
        self.prj_dropped = []
        self.projects_to_do = list(projects)

        while self.projects_to_do:
            p = self.projects_to_do.pop(0)
            # save the vs environment
            saved_env = copy.copy(self.vs_env)
            try:
                st = time.time()
                if self.__build_one(p):
                    self.prj_skipped.append(p.name)
                else:
                    msg = "%-*s (%.3f s)" % (
                        Project.name_len,
                        p.name,
                        time.time() - st,
                    )
                    self.prj_done.append(msg)
            except KeyboardInterrupt:
                traceback.print_exc()
                log.error_exit(f"Interrupted on {p.name}")
            except Exception:
                traceback.print_exc()
                log.end(mark_error=True)
                if self.opts.keep_going:
                    self.prj_err.append(p.name)
                    self._drop_proj(p)
                else:
                    log.error_exit(f"{p.name} build failed")
            self.vs_env = saved_env

        script_title(None)
        if self.prj_done:
            log.message("")
            log.message("Project(s) built:")
            for p in self.prj_done:
                log.message(f"    {p}")

        if self.prj_skipped:
            log.message("")
            log.message("Project(s) skipped (already built):")
            for p in self.prj_skipped:
                log.message(f"    {p}")

        if self.prj_err:
            log.message("")
            log.message("Project(s) not built:")
            for p in self.prj_err:
                log.message(f"    {p}")

            miss = len(self.prj_err)
            if self.prj_dropped:
                log.message("")
                log.message("Missing dependecies:")
                for p in self.prj_dropped:
                    log.message(f"    {p}")
                miss += len(self.prj_dropped)

            # Don't fool appveyor
            log.error_exit("%u project(s) missing ;(" % (miss,))

        log.close()

    def __prepare_build(self, projects):
        if not os.path.exists(self.working_dir):
            log.log(f"Creating working directory {self.working_dir}")
            os.makedirs(self.working_dir)

        shutil.copy(
            os.path.join(self.opts.patches_root_dir, "stack.props"), self.working_dir
        )

        build_dir = os.path.join(
            self.working_dir, "..", "..", "..", "gtk", self.opts.platform
        )
        if not os.path.exists(build_dir):
            log.log(f"Creating directory {build_dir}")
            os.makedirs(build_dir)

        if not os.path.exists(self.opts.export_dir):
            log.log(f"Creating export directory {self.opts.export_dir}")
            os.makedirs(self.opts.export_dir)

        script_title("* Downloading")
        log.start("Downloading packages")
        for p in projects:
            if self.__download_one(p):
                return True
        log.end()

        return False

    def _load_built_files(self):
        """Return a set with all the files present in the final, installation,
        dir."""
        return dirlist2set(self.gtk_dir)

    def __build_one(self, proj):
        """Build one project, return True if skipped."""
        # Check if we update the tarball / git
        proj.builder = self
        self.__project = proj
        proj.prepare_build_dir()
        proj.export()

        if self.opts.fast_build and not proj.clean:
            t = proj.mark_file_exist()
            if t:
                log.message(f"Fast build:skipping project {proj.name}, built @ {t}")
                proj.builder = None
                self.__project = None
                return True

        proj.mark_file_remove()
        log.start(f"Building project {proj.name} ({proj.version})")
        script_title(f"{proj.name} ({proj.version})")

        proj.pkg_dir = f"{proj.build_dir}-rel"
        shutil.rmtree(proj.pkg_dir, ignore_errors=True)
        os.makedirs(proj.pkg_dir)

        # Original path, converted to list
        paths = self.vs_env["PATH"].split(";")
        # Add the paths needed
        for d in proj.all_dependencies:
            t = d.get_path()
            if t:
                if isinstance(t, tuple):
                    # pre/post
                    if t[0]:
                        # Add at the beginning,
                        paths.insert(0, t[0])
                    if t[1]:
                        # Add at the end (msys, )
                        paths.append(t[1])
                else:
                    # Single path,  at the beginning
                    paths.insert(0, t)
            # Extra enviroment from the dependency, normally the tools ...
            d.apply_extra_env(self.vs_env)

        # Make the (eventually) new path
        self.vs_env["PATH"] = ";".join(paths)

        proj.patch()
        skip_deps = proj.build()

        log.debug(f"copying {proj.pkg_dir} to {self.gtk_dir}")
        self.copy_all(proj.pkg_dir, self.gtk_dir)
        shutil.rmtree(proj.pkg_dir, ignore_errors=True)

        proj.post_install()

        proj.builder = None
        self.__project = None

        if self.opts.make_zip:
            # Create file list
            cur = self._load_built_files()
            # delta with the old
            new = cur - self.file_built
            if new:
                self.__build_zip(proj, new, cur)
            else:
                log.log(f"{proj.name}:zip not needed (tool?)")

        # Drop the mark file for all the projects that depends on this so we rebuild them
        if not skip_deps:
            first = True
            for p in self.prj_to_mark:
                if proj in p.all_dependencies:
                    if first:
                        first = False
                        log.debug(f"Forcing build of {proj.name} dependent")
                    log.debug(f" > Mark {p.name} ...")
                    p.mark_file_remove()
                    self.prj_to_mark.remove(p)

        # Mark this project done correctly
        proj.mark_file_write()

        script_title(None)
        log.end()
        return False

    def __build_zip(self, proj, new, cur):
        t_ver = proj.version[4:] if proj.version.startswith("git/") else proj.version
        _t = [c if c.isalnum() else "_" for c in t_ver]
        ver_part = "".join(_t)

        zip_file = os.path.join(self.zip_dir, f"{proj.prj_dir}-{ver_part}")
        self.make_zip(zip_file, new)
        # use the current file set
        self.file_built = cur

    def make_zip(self, name, files):
        make_zip(name, files, skip_spc=len(self.gtk_dir))

    def make_dir(self, path):
        if not os.path.exists(path):
            os.makedirs(path)

    def copy_all(self, srcdir, destdir):
        self.make_dir(destdir)
        for f in glob.glob(f"{srcdir}\\*"):
            self.__copy_to(f, destdir)

    def __copy_to(self, src, destdir):
        self.make_dir(destdir)
        for f in glob.glob(src):
            if os.path.isdir(f):
                name = os.path.basename(f)
                dest_subdir = os.path.join(destdir, name)
                self.make_dir(dest_subdir)
                for item in os.listdir(f):
                    self.__copy_to(os.path.join(f, item), dest_subdir)
            else:
                shutil.copy2(f, destdir)

    def __copy(self, src, destdir):
        if os.path.isdir(src):
            dst = os.path.join(destdir, os.path.basename(src))
            log.debug(f"copying '{src}' to '{dst}'")
            shutil.copytree(src, dst)
        else:
            log.debug(f"copying '{src}' to '{destdir}'")
            shutil.copy(src, destdir)

    def __hashfile(self, file_name):
        hash_calc = hashlib.sha256()
        with open(file_name, "rb") as fi:
            for chunk in iter(lambda: fi.read(4096), b""):
                hash_calc.update(chunk)
        return hash_calc.hexdigest()

    def __check_hash(self, proj):
        if hasattr(proj, "hash"):
            hc = self.__hashfile(proj.archive_file)
            if hc != proj.hash:
                log.error_exit(
                    "Hash mismatch on %s:\n  Calculated '%s'\n  Expected   '%s'\n"
                    % (
                        proj.archive_file,
                        hc,
                        proj.hash,
                    )
                )
                return True

            # Print the correct hash
            if self.opts.check_hash:
                log.message(f"Hash ok on {proj.archive_file} ({hc})")
            else:
                log.debug(f"Hash ok on {proj.archive_file} ({hc})")
        return False

    def __download_progress(self, count, block_size, total_size):
        c_size = count * block_size
        if total_size > 0:
            # Percentage
            perc = (100 * c_size) // total_size
            if perc != self._old_perc:
                perc = min(perc, 100)
                self._old_perc = perc
                sp = "%s (%u k) - %u%%" % (
                    self._downloading_file,
                    total_size / 1024,
                    self._old_perc,
                )
                print(sp, end="\r")
                if len(sp) > self._old_print:
                    # Save the len to delete the line when we change file
                    self._old_print = len(sp)
        else:
            # Only the current, we don't know the size
            sp = "%s - %u k" % (
                self._downloading_file,
                c_size / 1024,
            )
            print(sp, end="\r")
            if len(sp) > self._old_print:
                self._old_print = len(sp)

    def urlretrieve(self, url, filename, reporthook, ssl_ignore_cert=False):
        """Retrieve a URL into a temporary location on disk.

        Requires a URL argument. If a filename is passed, it is used as
        the temporary file location. The reporthook argument should be
        a callable that accepts a block number, a read size, and the
        total file size of the URL target. The data argument should be
        valid URL encoded data.

        If a filename is passed and the URL points to a local resource,
        the result is a copy from local file to new file.

        Returns a tuple containing the path to the newly created
        data file as well as the resulting HTTPMessage object.
        """

        ssl_ctx = ssl._create_unverified_context() if ssl_ignore_cert else None
        msg = f"Opening {url} ..."
        print(msg, end="\r")
        with contextlib.closing(urlopen(url, None, context=ssl_ctx)) as fp:
            print(
                "%*s"
                % (
                    len(msg),
                    "",
                ),
                end="\r",
            )
            headers = fp.info()

            with open(filename, "wb") as tfp:
                read = 0
                blocknum = 0
                result = filename, headers
                size = (
                    int(headers["Content-Length"])
                    if "content-length" in headers
                    else -1
                )
                bs = 1024 * 8
                reporthook(blocknum, bs, size)

                while True:
                    block = fp.read(bs)
                    if not block:
                        break
                    read += len(block)
                    tfp.write(block)
                    blocknum += 1
                    reporthook(blocknum, bs, size)

        if size >= 0 and read < size:
            raise ContentTooShortError(
                "retrieval incomplete: got only %i out of %i bytes" % (read, size),
                result,
            )

        return result

    def __download_one(self, proj):
        if not proj.archive_file:
            log.debug(
                f"archive file is not specified for project {proj.name}, skipping"
            )
            return False

        if os.path.exists(proj.archive_file):
            log.debug(f"archive {proj.archive_file} already exists")
            return self.__check_hash(proj)

        if not os.path.exists(self.opts.archives_download_dir):
            log.log(
                f"Creating archives download directory {self.opts.archives_download_dir}"
            )

            os.makedirs(self.opts.archives_download_dir)

        log.start_verbose(f"Downloading {proj.archive_file}")
        # Setup for progress show
        self._downloading_file = proj.archive_file
        self._old_perc = -1
        self._old_print = 0
        try:
            self.urlretrieve(
                proj.archive_url, proj.archive_file, self.__download_progress
            )
        except (ssl.SSLError, URLError) as e:
            print(f"Exception downloading file '{proj.archive_url}'")
            print(e)
            if hasattr(proj, "hash"):
                self._old_perc = -1
                self._old_print = 0
                self.urlretrieve(
                    proj.archive_url,
                    proj.archive_file,
                    self.__download_progress,
                    ssl_ignore_cert=True,
                )
            else:
                print("Hash not present, bailing out ;(")
                raise
        log.end()

        print(
            "%-*s" % (self._old_print, f"{proj.archive_file} - Download finished"),
        )
        return self.__check_hash(proj)

    def __sub_vars(self, s):
        if "%" not in s:
            return s
        d = {
            "platform": self.opts.platform,
            "configuration": self.opts.configuration,
            "build_dir": self.opts.build_dir,
            "vs_ver": self.opts.vs_ver,
            "gtk_dir": self.gtk_dir,
            "vs_ver_year": self.vs_ver_year,
        }
        python = None
        if self.__project is not None:
            d["pkg_dir"] = self.__project.pkg_dir
            d["build_dir"] = self.__project.build_dir
            python = Path(sys.executable).parent
            d["python_dir"] = python
            # Add perl only if the project depends on them

            p = Project.get_project("perl")
            if p in self.__project.all_dependencies:
                perl = Project.get_tool_base_dir(p)
                d["perl_dir"] = perl

        d["msbuild_opts"] = self._create_msbuild_opts(python)
        return s % d

    def exec_vs(self, cmd, working_dir=None, add_path=None):
        self.__execute(
            self.__sub_vars(cmd),
            working_dir=working_dir,
            add_path=add_path,
            env=self.vs_env,
        )

    def exec_cargo(
        self, params="", working_dir=None, rustc_opts=None, rust_version="stable"
    ):
        cmd = "cargo"
        if self.opts.cargo_opts:
            cmd += f" {self.opts.cargo_opts}"
        if params:
            cmd += f" {params}"

        cargo_home = Project.get_tool_path("cargo")

        env = os.environ.copy()
        env["RUSTUP_HOME"] = cargo_home
        env["CARGO_HOME"] = cargo_home
        if rustc_opts is not None:
            env.update(rustc_opts)

        # set platform
        rustup = os.path.join(cargo_home, "rustup.exe")
        self.__execute(
            f'{rustup} default {rust_version}-{"i686" if self.x86 else "x86_64"}-pc-windows-msvc',
            env=env,
        )

        # build
        self.__execute(
            self.__sub_vars(cmd),
            working_dir=working_dir,
            add_path=cargo_home,
            env=self.vs_env,
        )

    def exec_cmd(self, cmd, working_dir=None, add_path=None):
        self.__execute(self.__sub_vars(cmd), working_dir=working_dir, add_path=add_path)

    def exec_ninja(self, params="", working_dir=None, add_path=None):
        cmd = "ninja"
        if self.opts.ninja_opts:
            cmd += f" {self.opts.ninja_opts}"
        if params:
            cmd += f" {params}"
        self.__execute(
            self.__sub_vars(cmd),
            working_dir=working_dir,
            add_path=add_path,
            env=self.vs_env,
        )

    def install(self, build_dir, pkg_dir, *args):
        if len(args) == 1:
            args = args[0].split()
        dest = os.path.join(pkg_dir, self.__sub_vars(args[-1]))
        self.make_dir(dest)
        for f in args[:-1]:
            src = os.path.join(self.__sub_vars(build_dir), self.__sub_vars(f))
            log.debug(f"copying {src} to {dest}")
            self.__copy_to(src, dest)

    def install_dir(self, build_dir, pkg_dir, src, dest):
        src = os.path.join(build_dir, self.__sub_vars(src))
        dest = os.path.join(pkg_dir, self.__sub_vars(dest))
        log.debug(f"copying {src} content to {dest}")
        self.copy_all(src, dest)

    def exec_msys(self, args, working_dir=None):
        self.__execute(
            args,
            working_dir=working_dir,
            add_path=os.path.join(self.opts.msys_dir, "usr", "bin"),
        )

    def __execute(self, args, working_dir=None, add_path=None, env=None):
        log.debug(f"running {args}, cwd={working_dir}, path+={add_path}")
        if add_path:
            env = dict(env) if env is not None else dict(os.environ)
            self.__add_path(env, add_path)
        if self.opts.capture_out:
            try:
                res = subprocess.run(
                    args,
                    cwd=working_dir,
                    env=env,
                    shell=True,
                    check=True,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.STDOUT,
                    text=True,
                    errors="ignore",
                )
            except subprocess.CalledProcessError as e:
                # Dump the lines that leads to the error
                log.messages_dump(e.stdout, f"Error building '{self.__project.name}'")
                # and let the caller handle this
                raise

            log.messages_dump(res.stdout, prt=self.opts.print_out)
        else:
            subprocess.check_call(args, cwd=working_dir, env=env, shell=True)

    def __add_path(self, env, folder):
        key = next((k for k in env if k.lower() == "path"), None)
        if key:
            env[key] = f"{env[key]};{folder}"
        else:
            key = "path"
            env[key] = folder
        log.debug(f"Changed path env variable to '{env[key]}'")



#  Copyright (C) 2016 The Gvsbuild Authors
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, see <http://www.gnu.org/licenses/>.

from gvsbuild.utils.base_project import Project, ProjectType, project_add
from gvsbuild.utils.simple_ui import log


@project_add
class DevShell(Project):
    def __init__(self):
        Project.__init__(
            self,
            "dev-shell",
            # We may need all tools
            dependencies=["tools"],
            version="0.1.0",
            # We don't want this project to be built with the group 'all'
            type=ProjectType.IGNORE,
        )
        self.meson = True

    def unpack(self):
        # Nothing to do, it's not really a project
        pass

    def export(self):
        # Nothing to do, it's not really a project
        pass

    def finalize_dep(self, builder, deps):
        if builder.opts.skip:
            for s in builder.opts.skip:
                p = Project.get_project(s)
                if p in deps:
                    log.log(f"dev-shell: skip {s}")
                    deps.remove(p)
                    if s in ["meson", "python"]:
                        # We disable the meson management
                        self.meson = False

    def build(self):
        # Do the shell
        print("")
        print("gvsbuild dev shell. Type exit to exit :)")
        print("")
        print("The environment var GTK_BASE_DIR points to the gtk installation dir")
        print(f"({self.builder.gtk_dir})")
        print("if you need it e.g. as a --prefix option")
        print("")
        if self.meson:
            # Add a _meson env to use it directly
            meson_path = Project.get_tool_path("meson")
            self.builder.mod_env("_MESON", f"python {meson_path}\\meson.py")
            print("If you need to use meson you can use the _MESON environment, e.g.")
            print("%_MESON% configure")
            print("")

        # If you need to use it as a --prefix in some build test ...
        self.builder.mod_env("GTK_BASE_DIR", self.builder.gtk_dir)
        self.builder.mod_env("PROMPT", "[ gvsbuild shell ] $P $G", subst=True)
        self.builder.exec_vs("cmd", working_dir=self.builder.working_dir)
